# import files
from .gaga_helpers import *
from .gaga_helpers_plot import *
from .gaga_helpers_pet import *
from .gaga_helpers_spect import *
from .gaga_functions import *
from .gaga import *
from .gaga_model import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import scipy
import torch
import numpy as np
import gaga_phsp as gaga

speed_of_light = scipy.constants.speed_of_light * 1000 / 1e9


def from_exit_pos_to_ideal_pos(x, params):
    """
    input : consider input key exit position X,Y,Z
    output: parametrize with ideal position = p - c x t x dir

    Input: PrePosition_X PreDirection_X TimeFromBeginOfEvent
    Output: IdealPosition_X (+idem)
    """

    # input and output keys
    keys = params["keys_list"]
    keys_out = [
        "KineticEnergy",
        "IdealPosition_X",
        "IdealPosition_Y",
        "IdealPosition_Z",
        "PreDirection_X",
        "PreDirection_Y",
        "PreDirection_Z",
        "TimeFromBeginOfEvent",
        "EventPosition_X",
        "EventPosition_Y",
        "EventPosition_Z",
        "EventDirection_X",
        "EventDirection_Y",
        "EventDirection_Z",
    ]

    # Step1: name the columns according to key
    p_ideal, d_exit, e_pos, e_dir = gaga.get_key_3d(
        x,
        keys,
        ["PrePosition_X", "PreDirection_X", "EventPosition_X", "EventDirection_X"],
    )
    t_exit = gaga.get_key(x, keys, ["TimeFromBeginOfEvent"])[0]
    ene = gaga.get_key(x, keys, ["KineticEnergy"])[0]

    # Step2: compute P_ideal
    P_ideal = p_ideal - speed_of_light * t_exit * d_exit

    # Step3: stack
    # x = torch.stack((ene, P_ideal, d_exit, t_exit, e_pos), dim=0).T # FIXME torch or numpy ?
    x = np.concatenate((ene, P_ideal, d_exit, t_exit, e_pos, e_dir), axis=1)

    # end
    return x, keys_out


def from_ideal_pos_to_exit_pos(x, params):

    # input and output keys
    keys = params["keys_list"]
    keys_out = [
        "KineticEnergy",
        "PrePosition_X",
        "PrePosition_Y",
        "PrePosition_Z",
        "PreDirection_X",
        "PreDirection_Y",
        "PreDirection_Z",
        "TimeFromBeginOfEvent",
        "EventPosition_X",
        "EventPosition_Y",
        "EventPosition_Z",
        "EventDirection_X",
        "EventDirection_Y",
        "EventDirection_Z",
    ]

    # Step1: name the columns according to key
    p_ideal, d_exit, e_pos, e_dir = gaga.get_key_3d(
        x,
        keys,
        ["IdealPosition_X", "PreDirection_X", "EventPosition_X", "EventDirection_X"],
    )
    t_exit = gaga.get_key(x, keys, ["TimeFromBeginOfEvent"])[0]
    ene = gaga.get_key(x, keys, ["KineticEnergy"])[0]

    # Step2: compute P_ideal
    P_exit = p_ideal + speed_of_light * t_exit * d_exit

    # Step3: stack
    # x = torch.stack((ene, P_ideal, d_exit, t_exit, e_pos), dim=0).T # FIXME torch or numpy ?
    x = np.concatenate((ene, P_exit, d_exit, t_exit, e_pos, e_dir), axis=1)

    # end
    return x, keys_out



import torch
import torch.nn as nn
from torch.autograd import grad as torch_grad
from torch import Tensor
import numpy as np

"""
Set of helpers functions for losses an penalties
"""


class WassersteinLoss(torch.nn.Module):
    """
    Wasserstein Loss
    """

    def __init__(self):
        super(WassersteinLoss, self).__init__()

    def forward(self, x, y):
        return torch.mean(x)


class WassersteinNegLoss(torch.nn.Module):
    """
    Wasserstein Loss
    """

    def __init__(self):
        super(WassersteinNegLoss, self).__init__()

    def forward(self, x, y):
        return -torch.mean(x)


class HingeLoss(torch.nn.Module):
    """
    Hinge Loss
    """

    def __init__(self):
        super(HingeLoss, self).__init__()

    def forward(self, x, y):
        return torch.mean(torch.nn.ReLU()(1.0 + x))


class HingeNegLoss(torch.nn.Module):
    """
    Hinge Loss
    """

    def __init__(self):
        super(HingeNegLoss, self).__init__()

    def forward(self, x, y):
        return torch.mean(torch.nn.ReLU()(1.0 - x))


def zero_penalty(self, real_data, fake_data):
    """
    No penalty
    """
    return 0


def clamp_parameters(self):
    """
    Clamp
    """
    for p in self.D.parameters():
        p.data.clamp_(self.clamp_lower, self.clamp_upper)


def get_interpolated_gradient(self, real_data, fake_data):
    """
    Common function to gradient penalty functions
    """

    alpha = torch.rand(self.batch_size, 1)  # , 1, 1)
    alpha = alpha.expand_as(real_data)
    alpha = alpha.to(self.current_gpu_device)

    # interpolated
    interpolated = alpha * real_data + (1 - alpha) * fake_data
    interpolated = (
        interpolated.clone().detach().requires_grad_(True).to(self.current_gpu_device)
    )

    # Calculate probability of interpolated examples
    prob_interpolated = self.D(interpolated)

    # gradient
    ones = torch.ones(prob_interpolated.size()).to(self.current_gpu_device)

    gradients = torch_grad(
        outputs=prob_interpolated,
        inputs=interpolated,
        grad_outputs=ones,
        create_graph=True,  # needed ?
        retain_graph=True,  # needed ?
        only_inputs=True,
    )[0]

    # gradients.requires_grad_(False) # FIXME
    return gradients


def gradient_penalty(self, real_data, fake_data):
    """
    Gulrajani2017 $(||\nabla_a D(a)||_2 - 1)^2$
    with a = interpolated samples.
    Called two-sided penalty.
    1-GP = one centered GP.
    """

    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)

    # norm
    # gradients_norm = torch.sqrt(torch.sum(gradients ** 2, dim=1) + 1e-12)
    gradients_norm = gradients.norm(2, dim=1)

    # Two sides penalty
    gradient_penalty = ((gradients_norm - 1) ** 2).mean()

    return gradient_penalty


def gradient_penalty_centered(self, real_data, fake_data):
    """
    Thanh-Tung2019 $(||\nabla_a D(a)||_2)^2$
    with a = interpolated samples.
    0-GP = zero centered GP.
    """

    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)

    # norm
    # gradients_norm = torch.sqrt(torch.sum(gradients ** 2, dim=1) + 1e-12)
    gradients_norm = gradients.norm(2, dim=1)

    # Two sides penalty
    gradient_penalty = ((gradients_norm) ** 2).mean()

    return gradient_penalty


def gradient_penalty_max(self, real_data, fake_data):
    """
    Petzka2018 $(max||\nabla_a D(a)||_2 - 1)^2$
    with a = interpolated samples.
    """

    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)

    # norm
    # gradients_norm = torch.sqrt(torch.sum(gradients ** 2, dim=1) + 1e-12)
    gradients_norm = gradients.norm(2, dim=1)

    # max FIXME why  **2 ????
    gradient_penalty = (
        torch.max(gradients_norm - 1, torch.zeros_like(gradients_norm)) ** 2
    ).mean()
    # ? gradient_penalty = torch.nn.ReLU()(gradients_norm - 1, torch.zeros_like(gradients_norm))**2).mean()

    return gradient_penalty


def gradient_penalty_linf_hinge(self, real_data, fake_data):
    """
    Jolicoeur2019 (TEST)
    """

    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)

    # L_inf margin
    gradients_norm = gradients.norm(1, dim=1)

    # hinge
    gradient_penalty = torch.nn.ReLU()(gradients_norm - 1).mean()

    return gradient_penalty


def gradient_penalty_linf_hinge_abs(self, real_data, fake_data):
    """
    Jolicoeur2019 (TEST), idem with abs
    """

    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)

    # abs gradients
    gradients = torch.abs(gradients)

    # L_inf margin
    gradients_norm = gradients.norm(1, dim=1)

    # hinge
    gradient_penalty = torch.nn.ReLU()(gradients_norm - 1).mean()

    return gradient_penalty


def GP_L1_LS(self, real_data, fake_data):
    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)
    # norm
    gradients_norm = gradients.norm(1, dim=1)
    # penalty
    gradient_penalty = (gradients_norm - 1).pow(2)
    return gradient_penalty.mean()


def GP_L2_LS(self, real_data, fake_data):
    # Gulrajani2017 $(||\nabla_a D(a)||_2 - 1)^2$
    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)
    # norm
    gradients_norm = gradients.norm(2, dim=1)
    # penalty
    gradient_penalty = (gradients_norm - 1).pow(2)
    return gradient_penalty.mean()


def GP_Linf_LS(self, real_data, fake_data):
    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)
    # norm
    gradients_norm = gradients.norm(p=float("inf"), dim=1)
    # inf norm is defined as max(sum(abs(x), axis=1)) for matrix and max(abs(x)) for vector
    # https://github.com/pytorch/pytorch/issues/24802
    # grad_abs = torch.abs(grad)  # Absolute value of gradient
    # grad_norm, _ = torch.max(grad_abs, 1)
    # penalty
    gradient_penalty = (gradients_norm - 1).pow(2)
    return gradient_penalty.mean()


def GP_L1_Hinge(self, real_data, fake_data):
    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)
    # norm
    gradients_norm = gradients.norm(1, dim=1)
    # penalty
    gradient_penalty = torch.nn.ReLU()(gradients_norm - 1)
    return gradient_penalty.mean()


def GP_L2_Hinge(self, real_data, fake_data):
    # Gulrajani2017 $(||\nabla_a D(a)||_2 - 1)^2$
    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)
    # norm
    gradients_norm = gradients.norm(2, dim=1)
    # penalty
    gradient_penalty = torch.nn.ReLU()(gradients_norm - 1)
    return gradient_penalty.mean()


def GP_Linf_Hinge(self, real_data, fake_data):
    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)
    # norm
    gradients_norm = gradients.norm(p=float("inf"), dim=1)
    # inf norm is defined as max(sum(abs(x), axis=1)) for matrix and max(abs(x)) for vector
    # https://github.com/pytorch/pytorch/issues/24802
    # grad_abs = torch.abs(grad)  # Absolute value of gradient
    # grad_norm, _ = torch.max(grad_abs, 1)
    # penalty
    gradient_penalty = torch.nn.ReLU()(gradients_norm - 1)
    return gradient_penalty.mean()


def GP_0GP(self, real_data, fake_data):
    # thanh-tung2019, zero-centered
    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)
    # norm
    gradients_norm = gradients.norm(2, dim=1)
    # penalty
    gradient_penalty = (gradients_norm).pow(2)
    return gradient_penalty.mean()


def GP_SquareHinge(self, real_data, fake_data):
    # petzka2018, square hinge
    # gradient
    gradients = get_interpolated_gradient(self, real_data, fake_data)
    # norm
    gradients_norm = gradients.norm(2, dim=1)
    # penalty
    gradient_penalty = (torch.nn.ReLU()(gradients_norm - 1)).pow(2)
    return gradient_penalty.mean()


def langevin_latent_sampling(G, D, params, z):
    print("generate with langevin")
    n = len(z)
    print("size", n)

    """
    line 85 run_synthetic.py
    line 49 langevin.py
    https://github.com/clear-nus/DGflow/blob/main/dgflow.py
    """
    # params
    n_steps = params["lgs_nb_steps"]
    eta = params["lgs_lr"]
    gamma = params["lgs_gamma"]
    f = params["lgs_f"]

    # parameters (will be in params)
    # n_steps = 20  # 50000 in che2020 ? ; 25 in ansari2020
    # alpha = 1.0  # for rejection
    # step_lr = 1e-4  ## see eta
    # eps_std = 1e-2
    # eta = 0.1  # step_lr = step_size
    # f = 'KL'  # KL logD JS ; used in "velocity"
    # gamma = 0.02  # diffusion term

    fn = None
    if f == "KL":
        fn = lambda d_score: torch.ones_like(d_score.detach())
    if f == "logD":
        fn = lambda d_score: 1 / (1 + d_score.detach().exp())
    if f == "JS":
        fn = lambda d_score: 1 / (1 + 1 / d_score.detach().exp())
    if not fn:
        print("ERROR lgs_f must be KL, LogD or JS")
        return z

    noise_factor = np.sqrt(gamma)
    print("langevin nb steps ", n_steps)
    print("langevin lr       ", eta)
    print("langevin gamma    ", gamma)
    print("langevin f        ", f)

    """
    KL-divergence by setting f = r log r 
    ==> equivalent to DDLS Che2020 ? NO 
    s = torch.ones_like(d_score.detach())
    
    f == 'logD' => s = 1 / (1 + d_score.detach().exp())
    Discriminator Optimal Transport (DOT) (Tanaka, 2019) ; with λ = 1/2 ? unsure 
    NO it is different 
    
    f == 'JS' => s = 1 / (1 + 1 / d_score.detach().exp())
    proposed by ansari2020 ?
    
    DGflow performs well even without the diffusion term (i.e., with γ = 0)
    """

    def velocity(z):
        z_t = z.clone()
        z_t.requires_grad_(True)
        if z_t.grad is not None:
            z_t.grad.zero_()
        x_t = G(z_t)
        d_score = D(x_t)
        # s = torch.ones_like(d_score.detach())
        # s = 1 / (1 + d_score.detach().exp())
        # s = 1 / (1 + 1 / d_score.detach().exp())  # FIXME use f
        s = fn(d_score)
        s.expand_as(z_t)
        d_score.backward(torch.ones_like(d_score).to(z.device))
        grad = z_t.grad
        return s.data * grad.data

    # loop on number of steps
    for i in range(n_steps):
        print(i, len(z))
        v = velocity(z)
        # FIXME warning randn !
        z = z.data + eta * v + np.sqrt(2 * eta) * noise_factor * torch.randn_like(z)
        """
            v = velocity(z)  <--- grad ?
            z = z.data + eta * v + np.sqrt(2*eta) * noise_factor * torch.randn_like(z)
        """
        """
            eps = eps_std * xp.random.randn(batch_size, z_dim).astype(xp.float32)
            E, grad = e_grad(z, P, gen, dis, alpha, ret_e=True)
            z = z - step_lr * grad[0] + eps
        """

    return z



import numpy as np
import torch
from torch import Tensor
import gaga_phsp as gaga
import datetime
import time
import garf
from garf.helpers import get_gpu_device
import gatetools.phsp as phsp
from scipy.stats import entropy
from scipy.spatial.transform import Rotation
import itk
import logging
import sys
import os
from box import Box, BoxList

logger = logging.getLogger(__name__)

"""
date format
"""
date_format = "%Y-%m-%d %H:%M:%S"


def update_params_with_user_options(params, user_param):
    """
    Update the dict 'params' with the options set by the user on the command line
    """
    for up in user_param:
        params[up[0]] = up[1]


def param_check_keys(params, read_keys):
    """
    Consider the 'keys' tag in param
    - if not exist consider read_keys
    - remove '_' if needed
    - convert to keys_list
    - check keys
    """

    # if no keys, just consider the read ones
    if "keys" not in params:
        params["keys"] = read_keys.join(" ")

    # on command line, '#' may be used to replace space, so we put it back
    if "#" in params["keys"]:
        params["keys"] = params["keys"].replace("#", " ")

    # build the list of keys
    params.keys_list = phsp.str_keys_to_array_keys(params["keys"])

    # check
    for k in params.keys_list:
        if k not in read_keys:
            print(f'Error, the key "{k}" does not belong to read keys: {read_keys}')
            exit(0)

    # debug
    # print('keys:      ', params['keys'])
    # print('keys_list: ', params.keys_list)


def check_input_params(params, fatal_on_unknown_keys=True):
    required = [
        "gpu_mode",
        "model",
        "d_layers",
        "g_layers",
        "d_learning_rate",
        "g_learning_rate",
        "optimiser",
        "d_nb_update",
        "g_nb_update",
        "loss",
        "penalty",
        "penalty_weight",
        "batch_size",
        "epoch",
        "d_dim",
        "g_dim",
        "z_dim",
        "r_instance_noise_sigma",
        "f_instance_noise_sigma",
        "activation",
        "z_rand_type",
        "shuffle",
        "keys",
        "epoch_dump",
        "keys_list",
    ]
    automated = [
        "params_filename",
        "training_size",
        "x_dim",
        "progress_bar",
        "training_filename",
        "start_date",
        "hostname",
        "d_nb_weights",
        "g_nb_weights",
        "x_mean",
        "x_std",
        "end_epoch",
        "Duration",
        "duration",
        "end_date",
        "output_filename",
        "cond_keys",
        "current_gpu_mode",
        "current_gpu_device",
    ]

    # forced
    if "r_instance_noise_sigma" not in params:
        params["r_instance_noise_sigma"] = -1
    if "f_instance_noise_sigma" not in params:
        params["f_instance_noise_sigma"] = -1

    """ 
        WARNING: management of the list of keys
        - user defined value in json is 'keys', but this is a reserve keyword :(
        - also user provide a simple str with space between the keys
        -> we convert into a list of str, named keys_list
        
        When the gan is trained, the params saved in the gan file is already a list. 
        
        Same for cond_keys.
    """

    # automated cond keys
    if "cond_keys" not in params:
        params["cond_keys"] = []
    else:
        if type(params["cond_keys"]) != list and type(params["cond_keys"]) != BoxList:
            params["cond_keys"] = phsp.str_keys_to_array_keys(params["cond_keys"])

    # keys (for backward compatible)
    if "keys_list" not in params:
        if type(params["keys"]) == list:
            params["keys_list"] = params["keys"].copy()
        else:
            params["keys_list"] = phsp.str_keys_to_array_keys(params["keys"])

    # old versions does not have some tags
    if "activation" not in params:
        params["activation"] = "relu"
    if "loss" not in params:
        params["loss"] = "wasserstein"
    if "model" not in params:
        params["model"] = "v3"
    if "penalty" not in params:
        params["penalty"] = "clamp"
    if "penalty_weight" not in params:
        params["penalty_weight"] = 0
    if "z_rand_type" not in params:
        params["z_rand_type"] = "randn"
    if "epoch_dump" not in params:
        params["epoch_dump"] = -1

    # check required
    for req in required:  # + automated
        if req not in params:
            print(f'Error, the parameters "{req}" is required in {params}')
            exit(0)

    # look unknown param
    optional = [
        "start_pth",
        "start_epoch",
        "schedule_learning_rate_step",
        "schedule_learning_rate_gamma",
        "label_smoothing",
        "spectral_norm",
        "epoch_store_model_every",
        "RMSprop_d_momentum",
        "RMSprop_g_momentum",
        "RMSProp_d_alpha",
        "RMSProp_g_alpha",
        "GAN_model",
        "RMSProp_d_weight_decay",
        "RMSProp_g_weight_decay",
        "RMSProp_d_centered",
        "RMSProp_g_centered",
    ]
    for p in params:
        if p[0] == "#":
            continue
        if p in optional:
            # print('Found optional: ', p)
            pass
        else:
            if p not in required + automated:
                print(
                    f'Warning unknown key named "{p}" in the parameters (deprecated?)'
                )
                if fatal_on_unknown_keys:
                    exit(0)

    # special for adam
    if params["optimiser"] == "adam":
        required_adam = ["g_weight_decay", "d_weight_decay", "beta_1", "beta_2"]
        for req in required_adam:
            if req not in params:
                print(f'Error, the Adam parameters "{req}" is required in {params}')
                exit(0)


def normalize_data(x):
    """
    Consider the input vector mean and std and normalize it
    """
    x_mean = np.mean(x, 0, keepdims=True)
    x_std = np.std(x, 0, keepdims=True)
    x = (x - x_mean) / x_std
    return x, x_mean, x_std


def print_pytorch_gpu_info(current_gpu_mode):
    print("pytorch version", torch.__version__)
    if current_gpu_mode == "cuda":
        print("GPU is enabled")
        print("CUDA version         ", torch.version.cuda)
        print("CUDA device counts   ", torch.cuda.device_count())
        print("CUDA current device  ", torch.cuda.current_device())
        n = torch.cuda.current_device()
        print("CUDA device name     ", torch.cuda.get_device_name(n))
        print("CUDA device address  ", torch.cuda.device(n))
    elif current_gpu_mode == "mps":
        print("GPU is enabled")
        print("MPS is used")
    elif current_gpu_mode == "cpu":
        print("CPU only (no GPU)")
    else:
        print('Error with current_gpu_mode, must be "cpu", "cuda" or "mps"')
        exit(-1)


def print_network(net):
    """
    Print info about a network
    """
    num_params = get_network_nb_parameters(net)
    print(net)
    print("Total number of parameters: %d" % num_params)


def get_network_nb_parameters(net):
    """
    Compute total nb of parameters
    """
    num_params = 0
    for param in net.parameters():
        num_params += param.numel()
    return num_params


def print_info(params, optim):
    """
    Print info about a trained GAN-PHSP
    """
    # print parameters
    for e in sorted(params):
        if (e[0] != "#") and (e != "x_mean") and (e != "x_std"):
            print("   {:22s} {}".format(e, str(params[e])))

    # additional info
    try:
        start = datetime.datetime.strptime(params["start date"], gaga.date_format)
        end = datetime.datetime.strptime(params["end date"], gaga.date_format)
    except:
        start = 0
        end = 0
    delta = end - start
    print("   {:22s} {}".format("Duration", delta))

    d_loss_real = np.asarray(optim["d_loss_real"][-1])
    d_loss_fake = np.asarray(optim["d_loss_fake"][-1])
    d_loss = d_loss_real + d_loss_fake
    g_loss = np.asarray(optim["g_loss"][-1])
    print("   {:22s} {}".format("Final d_loss", d_loss))
    print("   {:22s} {}".format("Final g_loss", g_loss))
    if "d_best_loss" in optim:
        print("   {:22s} {}".format("d_best_loss", optim["d_best_loss"]))
        print("   {:22s} {}".format("d_best_epoch", optim["d_best_epoch"]))

    # version
    p = "python"
    v = sys.version.replace("\n", "")
    print(f"   {p:22s} {v}")
    p = "pytorch"
    print(f"   {p:22s} {torch.__version__}")


def print_info_short(params, optim):
    p = Box(params)
    s = (
        f"H {p.params_filename} {p.d_dim} {p.g_dim} L {p.d_layers} {p.g_layers} Z {p.z_dim} "
        f"{p.penalty} {p.penalty_weight} lr {p.d_learning_rate} {p.g_learning_rate} "
    )
    try:
        s += f"sc {p.schedule_learning_rate_step} {p.schedule_learning_rate_gamma} "
    except:
        pass
    s += f"D:G {p.d_nb_update}:{p.g_nb_update} {p.epoch} {p.batch_size} {p.duration}"

    print(s)


def create_G_and_D_model(params):
    G = None
    D = None
    if params["model"] == "v3":
        G = gaga.Generator(params)
        D = gaga.Discriminator(params)
        return G, D
    if not D or not G:
        print("Error in create G and D model, unknown model version?")
        print(params["model"])
        print(params["GAN_model"])
        exit(0)


def auto_output_filename(params, output, output_folder):
    if output != "auto":
        params.output_filename = output
        return
    # create an output name with some params
    b, extension = os.path.splitext(os.path.basename(params.params_filename))
    output = f"{b}_{params.penalty}_{params.penalty_weight}_{params.end_epoch}.pth"
    params.output_filename = os.path.join(output_folder, output)


def load(
    filename, gpu_mode="auto", verbose=False, epoch=-1, fatal_on_unknown_keys=True
):
    """
    Load a GAN-PHSP
    Output params   = dict with all parameters
    Output G        = Generator network
    Output optim    = dict with information of the training process
    """

    current_gpu_mode, current_gpu_device = get_gpu_device(gpu_mode)
    nn = torch.load(filename, map_location=current_gpu_device)

    # get elements
    params = nn["params"]
    params["current_gpu_mode"] = current_gpu_mode
    params["current_gpu_device"] = current_gpu_device

    gaga.check_input_params(params, fatal_on_unknown_keys)
    if not "optim" in nn:
        optim = nn["model"]  ## FIXME compatibility --> to remove
    else:
        optim = nn["optim"]

    D_state = nn["d_model_state"]
    if epoch == -1:
        G_state = nn["g_model_state"]
    else:
        try:
            index = optim["current_epoch"].index(epoch)
        except:
            print(f"Epoch {epoch} is not in the list : {optim['current_epoch']}")
            exit(0)
        G_state = optim["g_model_state"][index]

    # create the Generator and the Discriminator (Critic)
    G, D = create_G_and_D_model(params)

    G.to(current_gpu_device)
    D.to(current_gpu_device)

    G.load_state_dict(G_state)
    D.load_state_dict(D_state)

    return params, G, D, optim


def get_min_max_constraints(params):
    """
    Compute the min/max values per dimension according to params['keys'] and params['constraints']
    """

    # clamp take normalisation into account
    x_dim = params["x_dim"]
    keys = params["keys"]
    ckeys = params["constraints"]
    cmin = np.ones((1, x_dim)) * -9999  # FIXME min value
    cmax = np.ones((1, x_dim)) * 9999  # FIXME max value
    for k, v in ckeys.items():
        try:
            index = keys.index(k)
            cmin[0, index] = v[0]
            cmax[0, index] = v[1]
        except:
            continue

    x_std = params["x_std"]
    x_mean = params["x_mean"]

    cmin = (cmin - x_mean) / x_std
    cmax = (cmax - x_mean) / x_std

    return cmin, cmax


def get_RMSProp_optimisers(self, p):
    """
    momentum (float, optional) – momentum factor (default: 0)
    alpha (float, optional) – smoothing constant (default: 0.99)
    centered (bool, optional) – if True, compute the centered RMSProp,
             the gradient is normalized by an estimation of its variance
    weight_decay (float, optional) – weight decay (L2 penalty) (default: 0)
    """

    d_learning_rate = p["d_learning_rate"]
    g_learning_rate = p["g_learning_rate"]

    if "RMSprop_d_momentum" not in p:
        p["RMSprop_d_momentum"] = 0
    if "RMSprop_g_momentum" not in p:
        p["RMSprop_g_momentum"] = 0

    if "RMSProp_d_alpha" not in p:
        p["RMSProp_d_alpha"] = 0.99
    if "RMSProp_g_alpha" not in p:
        p["RMSProp_g_alpha"] = 0.99

    if "RMSProp_d_weight_decay" not in p:
        p["RMSProp_d_weight_decay"] = 0
    if "RMSProp_g_weight_decay" not in p:
        p["RMSProp_g_weight_decay"] = 0

    if "RMSProp_d_centered" not in p:
        p["RMSProp_d_centered"] = False
    if "RMSProp_g_centered" not in p:
        p["RMSProp_g_centered"] = False

    RMSprop_d_momentum = p["RMSprop_d_momentum"]
    RMSprop_g_momentum = p["RMSprop_g_momentum"]
    RMSProp_d_alpha = p["RMSProp_d_alpha"]
    RMSProp_g_alpha = p["RMSProp_g_alpha"]
    RMSProp_d_weight_decay = p["RMSProp_d_weight_decay"]
    RMSProp_g_weight_decay = p["RMSProp_g_weight_decay"]
    RMSProp_d_centered = p["RMSProp_d_centered"]
    RMSProp_g_centered = p["RMSProp_g_centered"]

    d_optimizer = torch.optim.RMSprop(
        self.D.parameters(),
        lr=d_learning_rate,
        momentum=RMSprop_d_momentum,
        alpha=RMSProp_d_alpha,
        weight_decay=RMSProp_d_weight_decay,
        centered=RMSProp_d_centered,
    )
    g_optimizer = torch.optim.RMSprop(
        self.G.parameters(),
        lr=g_learning_rate,
        momentum=RMSprop_g_momentum,
        alpha=RMSProp_g_alpha,
        weight_decay=RMSProp_g_weight_decay,
        centered=RMSProp_g_centered,
    )

    return d_optimizer, g_optimizer


def get_z_rand(params):
    if "z_rand_type" in params:
        if params["z_rand_type"] == "rand":
            return torch.rand
        if params["z_rand_type"] == "randn":
            return torch.randn
    if "z_rand" in params:
        if params["z_rand"] == "uniform":
            return torch.rand
        if params["z_rand"] == "normal":
            return torch.randn
    params["z_rand_type"] = "randn"
    return torch.randn


def generate_samples2(
    params,
    G,
    D,
    n,
    batch_size=-1,
    normalize=False,
    to_numpy=False,
    z=None,
    cond=None,
    silence=False,
):
    # batch size -> if n is lower, batch size is n
    batch_size = int(batch_size)
    if batch_size == -1:
        batch_size = int(n)
        to_numpy = True

    if batch_size > n:
        batch_size = int(n)

    # get z random (gauss or uniform)
    z_rand = get_z_rand(params)

    # is this a conditional GAN ?
    is_conditional = not cond is None

    # normalize the input condition
    ncond = 0
    if is_conditional:
        # normalize the conditional vector
        xmean = params["x_mean"][0]
        xstd = params["x_std"][0]
        xn = params["x_dim"]
        cn = len(params["cond_keys"])
        ncond = cn
        # mean and std for cond only
        xmeanc = xmean[xn - cn : xn]
        xstdc = xstd[xn - cn : xn]
        # mean and std for non cond
        xmeannc = xmean[0 : xn - cn]
        xstdnc = xstd[0 : xn - cn]
        # normalize the condition
        cond = (cond - xmeanc) / xstdc
    else:
        if len(params["cond_keys"]) > 0:
            print(
                f'Error : GAN is conditional, you should provide the condition: {params["cond_keys"]}'
            )
            exit(0)

    langevin_latent_sampling_flag = False
    if "langevin_latent_sampling" in params:
        langevin_latent_sampling_flag = True

    m = 0
    z_dim = params["z_dim"]
    x_dim = params["x_dim"]
    rfake = np.empty((0, x_dim - ncond))
    device = params["current_gpu_device"]
    current_gpu_mode = params["current_gpu_mode"]
    while m < n:
        if not silence:
            print(f"Batch {m}/{n}")
        # no more samples than needed
        current_gpu_batch_size = batch_size
        if current_gpu_batch_size > n - m:
            current_gpu_batch_size = n - m
        # print('(G) current_gpu_batch_size', current_gpu_batch_size)

        # (checking Z allow to reuse z for some special test case)
        # if None == z:
        z = Tensor(z_rand(current_gpu_batch_size, z_dim)).to(device)

        # condition ?
        if is_conditional:
            if current_gpu_mode == "mps":
                # print("With device mps (gpu), convert data to float32")
                acond = cond[m : m + current_gpu_batch_size].astype(np.float32)
            else:
                acond = cond[m : m + current_gpu_batch_size]
            condx = (
                # Tensor(torch.from_numpy(cond[m : m + current_gpu_batch_size]))
                Tensor(torch.from_numpy(acond))
                .to(device)
                .view(current_gpu_batch_size, cn)
            )
            z = torch.cat((z.float(), condx.float()), dim=1)

        # FIXME test langevin
        if langevin_latent_sampling_flag:
            z = gaga.langevin_latent_sampling(G, D, params, z)

        fake = G(z)
        # put back to cpu to allow concatenation
        fake = fake.cpu().data.numpy()
        rfake = np.concatenate((rfake, fake), axis=0)

        m = m + current_gpu_batch_size

    if not normalize:
        x_mean = params["x_mean"]
        x_std = params["x_std"]
        if is_conditional:
            # do not consider the mean/std of the condition part
            x_mean = xmeannc
            x_std = xstdnc
        rfake = (rfake * x_std) + x_mean

    if to_numpy:
        return rfake

    return Tensor(torch.from_numpy(rfake)).to(device)


def generate_samples3(params, G, n, cond):
    """
    Like generate_samples2 but with less options, to see if it can be faster

    - batch size is managed elsewhere

    """
    # normalize the conditional vector
    xmean = params["x_mean"][0]
    xstd = params["x_std"][0]
    xn = params["x_dim"]
    cn = len(params["cond_keys"])
    ncond = cn
    # mean and std for cond only
    xmeanc = xmean[xn - cn : xn]
    xstdc = xstd[xn - cn : xn]
    # mean and std for non cond
    xmeannc = xmean[0 : xn - cn]
    xstdnc = xstd[0 : xn - cn]
    # normalize the condition
    cond = (cond - xmeanc) / xstdc

    m = 0
    z_dim = params["z_dim"]
    x_dim = params["x_dim"]
    rfake = np.empty((0, x_dim - ncond))

    # (checking Z allow to reuse z for some special test case)
    # if None == z:
    device = params["current_gpu_device"]
    z = Tensor(torch.randn(n, z_dim)).to(device)

    # condition ?
    condx = Tensor(torch.from_numpy(cond[m : m + n])).to(device).view(n, cn)
    z = torch.cat((z.float(), condx.float()), dim=1)

    # Go !!!
    fake = G(z)

    # put back to cpu to allow concatenation
    fake = fake.cpu().data.numpy()
    rfake = np.concatenate((rfake, fake), axis=0)

    # do not consider the mean/std of the condition part
    x_mean = xmeannc
    x_std = xstdnc
    rfake = (rfake * x_std) + x_mean

    return rfake


def Jensen_Shannon_divergence(x, y, bins, margin=0):
    # margin = 0#.01 # 5%
    r = [np.amin(x), np.amax(x)]
    if r[0] < 0:
        r = [r[0] + margin * r[0], r[1] + margin * r[1]]
    else:
        r = [r[0] - margin * r[0], r[1] + margin * r[1]]
    P, bin_edges = np.histogram(x, range=r, bins=bins, density=True)
    Q, bin_edges = np.histogram(y, range=r, bins=bins, density=True)

    _P = P / np.linalg.norm(P, ord=1)
    _Q = Q / np.linalg.norm(Q, ord=1)
    _M = 0.5 * (_P + _Q)
    return 0.5 * (entropy(_P, _M) + entropy(_Q, _M))


def sliced_wasserstein(x, y, l, p=1):
    l = int(l)
    ndim = len(x[0])

    if ndim == 1:
        d = wasserstein1D(x, y, p)
        d = d.data.cpu().numpy()
        return d

    l_batch_size = int(1e2)
    l_current = 0
    d = 0
    while l_current < l:
        # directions: matrix [ndim X l]
        directions = np.random.randn(ndim, l_batch_size)
        directions /= np.linalg.norm(directions, axis=0)

        # send to gpu if possible
        directions = torch.from_numpy(directions).to(device)

        # Projection (Radon) x = [n X ndim], px = [n X L]
        px = torch.matmul(x, directions)
        py = torch.matmul(y, directions)

        # sum wasserstein1D over all directions
        for i in range(l_batch_size):
            lx = px[:, i]
            ly = py[:, i]
            d += wasserstein1D(lx, ly, p)

        l_current += l_batch_size
        if l_current + l_batch_size > l:
            l_batch_size = l - l_current

    d = torch.pow(d / l, 1 / p)
    d = d.data.cpu().numpy()
    return d


def wasserstein1D(x, y, p=1):
    sx, indices = torch.sort(x)
    sy, indices = torch.sort(y)
    z = sx - sy
    return torch.sum(torch.pow(torch.abs(z), p)) / len(z)


def init_plane(n, angle, radius):
    """
    plane_U, plane_V, plane_point, plane_normal
    """

    n = int(n)
    logger.info(f"Initialisation of plane with radius {radius} ")
    plane_U = np.array([1, 0, 0])
    plane_V = np.array([0, 1, 0])
    r = Rotation.from_euler("y", angle, degrees=True)
    plane_U = r.apply(plane_U)
    plane_V = r.apply(plane_V)

    # normal vector is the cross product of two direction vectors on the plane
    plane_normal = np.cross(plane_U, plane_V)
    plane_normal = np.array([plane_normal] * n)

    center = np.array([0, 0, -radius])
    center = np.array([0, 0, -radius])
    center = r.apply(center)
    plane_center = np.array(
        [
            center,
        ]
        * n
    )

    plane = {
        "plane_U": plane_U,
        "plane_V": plane_V,
        "rotation": r,
        "plane_normal": plane_normal,
        "plane_center": plane_center,
    }
    # logger.info(f'Initialisation of plane {plane} ')
    return plane


def init_plane2(n, angle, radius, spect_table_shift_mm):
    """
    plane_U, plane_V, plane_point, plane_normal
    """

    n = int(n)
    plane_U = np.array([1, 0, 0])
    plane_V = np.array([0, 1, 0])
    r1 = Rotation.from_euler("x", 90, degrees=True)
    r2 = Rotation.from_euler("z", angle, degrees=True)
    r = r2 * r1
    plane_U = r.apply(plane_U)
    plane_V = r.apply(plane_V)

    # normal vector is the cross product of two direction vectors on the plane
    plane_normal = np.cross(plane_U, plane_V)
    plane_normal = np.array([plane_normal] * n)

    center = np.array([0, -spect_table_shift_mm, -radius])
    center = r.apply(center)
    plane_center = np.array([center] * n)

    plane = {
        "plane_U": plane_U,
        "plane_V": plane_V,
        "rotation": r.inv(),  # [r] * n,
        "plane_normal": plane_normal,
        "plane_center": plane_center,
    }

    return plane


def project_on_plane(x, plane, image_plane_size_mm, debug=False):
    """
    Project the x points (Ekine X Y Z dX dY dZ)
    on the image plane defined by plane_U, plane_V, plane_center, plane_normal
    """

    logger.info(f"Projection of {len(x)} particles on the plane")
    logger.info(f"Plane size is {image_plane_size_mm} mm")

    # shorter variable names

    # n is the normal plane, duplicated n times
    n = plane["plane_normal"][0 : len(x)]

    # c0 is the center of the plane, duplicated n times
    c0 = plane["plane_center"][0 : len(x)]

    # r is the rotation matrix of the plane, according to the current rotation angle (around Y)
    r = plane["rotation"][0 : len(x)]

    # p is the set of points position generated by the GAN
    p = x[:, 1:4]

    # u is the set of points direction generated by the GAN
    u = x[:, 4:7]

    # w is the set of vectors from all points to the plane center
    w = p - c0

    # project to plane
    ## dot product : out = (x*y).sum(-1)
    # https://rosettacode.org/wiki/Find_the_intersection_of_a_line_with_a_plane#Python
    # http://geomalgorithms.com/a05-_intersect-1.html
    # https://github.com/pytorch/pytorch/issues/18027
    ndotu = (n * u).sum(-1)  # dot product between normal plane (n) and direction (u)
    si = (
        -(n * w).sum(-1) / ndotu
    )  # dot product between normal plane and vector from plane to point (w)

    # only positive (direction to the plane)
    mask = si > 0
    mw = w[mask]
    mu = u[mask]
    mc0 = c0[mask]
    mn = n[mask]
    mx = x[mask]
    mp = p[mask]
    msi = si[mask]
    mnb = len(msi)
    logger.info(f"Remove negative direction, remains {mnb}/{len(x)}")

    # si is a (nb) size vector, expand it to (nb x 3)
    msi = np.array([msi] * 3).T

    # intersection between point-direction and plane
    psi = mp + msi * mu

    # apply the inverse of the rotation
    ri = r.inv()
    psip = ri.apply(psi)  # - offset

    # remove out of plane (needed ??)
    sizex = image_plane_size_mm[0] / 2.0
    sizey = image_plane_size_mm[1] / 2.0
    mask1 = psip[:, 0] < sizex
    mask2 = psip[:, 0] > -sizex
    mask3 = psip[:, 1] < sizey
    mask4 = psip[:, 1] > -sizey
    m = mask1 & mask2 & mask3 & mask4
    psip = psip[m]
    psi = psi[m]
    mp = mp[m]
    mu = mu[m]
    mx = mx[m]
    mc0 = mc0[m]
    nb = len(psip)
    logger.info(f"Remove points that are out of detector, remains {nb}/{len(x)}")

    # reshape results
    pu = psip[:, 0].reshape((nb, 1))  # u
    pv = psip[:, 1].reshape((nb, 1))  # v
    y = np.concatenate((pu, pv), axis=1)

    # rotate direction according to the plane
    mup = ri.apply(mu)
    norm = np.linalg.norm(mup, axis=1, keepdims=True)
    mup = mup / norm
    dx = mup[:, 0]
    dy = mup[:, 1]

    # FIXME -> clip arcos -1;1 ?

    # convert direction into theta/phi
    # theta is acos(dy)
    # phi is acos(dx)
    theta = np.degrees(np.arccos(dy)).reshape((nb, 1))
    phi = np.degrees(np.arccos(dx)).reshape((nb, 1))
    y = np.concatenate((y, theta), axis=1)
    y = np.concatenate((y, phi), axis=1)

    # concat the E
    E = mx[:, 0].reshape((nb, 1))
    data = np.concatenate((y, E), axis=1)

    return data


def project_on_plane2(x, plane, image_plane_size_mm):
    """
    Project the x points (Ekine X Y Z dX dY dZ)
    on the image plane defined by plane_U, plane_V, plane_center, plane_normal
    """

    # n is the normal plane, duplicated n times
    n = plane["plane_normal"][0 : len(x)]

    # c0 is the center of the plane, duplicated n times
    c0 = plane["plane_center"][0 : len(x)]

    # r is the rotation matrix of the plane, according to the current rotation angle (around Y)
    r = plane["rotation"]  # [0: len(x)]

    # p is the set of points position generated by the GAN
    p = x[:, 1:4]  # FIXME indices of the position

    # u is the set of points direction generated by the GAN
    u = x[:, 4:7]  # FIXME indices of the position

    # w is the set of vectors from all points to the plane center
    w = p - c0

    # project to plane
    # dot product : out = (x*y).sum(-1)
    # https://rosettacode.org/wiki/Find_the_intersection_of_a_line_with_a_plane#Python
    # http://geomalgorithms.com/a05-_intersect-1.html
    # https://github.com/pytorch/pytorch/issues/18027

    # dot product between normal plane (n) and direction (u)
    ndotu = (n * u).sum(-1)

    # dot product between normal plane and vector from plane to point (w)
    si = -(n * w).sum(-1) / ndotu

    # only positive (direction to the plane)
    mask = si > 0
    mu = u[mask]
    mc0 = c0[mask]
    mx = x[mask]
    mp = p[mask]
    msi = si[mask]
    mnb = len(msi)
    # print(f"Remove negative direction, remains {mnb}/{len(x)}")

    # si is a (nb) size vector, expand it to (nb x 3)
    msi = np.array([msi] * 3).T

    # intersection between point-direction and plane
    psi = mp + msi * mu

    # offset of the head
    psi = psi + c0[: len(psi)]

    # apply the inverse of the rotation
    psip = r.apply(psi)

    # remove out of plane (needed ??)
    sizex = image_plane_size_mm[0] / 2.0
    sizey = image_plane_size_mm[1] / 2.0
    mask1 = psip[:, 0] < sizex
    mask2 = psip[:, 0] > -sizex
    mask3 = psip[:, 1] < sizey
    mask4 = psip[:, 1] > -sizey
    m = mask1 & mask2 & mask3 & mask4
    psip = psip[m]
    mu = mu[m]
    mx = mx[m]
    nb = len(psip)
    # print(f"Remove points that are out of detector, remains {nb}/{len(x)}")

    # reshape results
    pu = psip[:, 0].reshape((nb, 1))  # u
    pv = psip[:, 1].reshape((nb, 1))  # v
    y = np.concatenate((pu, pv), axis=1)

    # rotate direction according to the plane
    mup = r.apply(mu)
    norm = np.linalg.norm(mup, axis=1, keepdims=True)
    mup = mup / norm
    dx = mup[:, 0]
    dy = mup[:, 1]

    # FIXME -> clip arcos -1;1 ?

    # convert direction into theta/phi
    # theta is acos(dy)
    # phi is acos(dx)
    theta = np.degrees(np.arccos(dy)).reshape((nb, 1))
    phi = np.degrees(np.arccos(dx)).reshape((nb, 1))
    y = np.concatenate((y, theta), axis=1)
    y = np.concatenate((y, phi), axis=1)

    # concat the E
    E = mx[:, 0].reshape((nb, 1))
    data = np.concatenate((y, E), axis=1)

    return data


def gaga_garf_generate_image(p):
    # param
    gan_params = p["gan_params"]
    G = p["G"]
    D = p["D"]
    batch_size = p["batch_size"]
    gan_batch_size = p["gan_batch_size"]
    plane = p["plane"]
    image_plane_size_mm = p["image_plane_size_mm"]
    debug = p["debug"]
    garf_nn = p["garf_nn"]
    garf_model = p["garf_model"]
    garf_param = p["garf_param"]
    pbar = p["pbar"]
    n = p["n"]

    ev = 0
    images = []
    sq_images = []
    while ev < n:
        # check generation of the exact nb of samples
        current_batch_size = batch_size
        if current_batch_size > n - ev:
            current_batch_size = n - ev

        # Step 1 : GAN
        t1 = time.time()
        logger.info(f"Generating {current_batch_size} events")
        x = gaga.generate_samples2(
            gan_params,
            G,
            D,
            current_batch_size,
            gan_batch_size,
            normalize=False,
            to_numpy=True,
        )
        # print('batch / x', current_batch_size, len(x))
        logger.info("Computation time: {0:.3f} sec".format(time.time() - t1))

        # Step 2 : Projection
        t1 = time.time()
        px = gaga.project_on_plane(
            x, plane, image_plane_size_mm=image_plane_size_mm, debug=debug
        )
        logger.info("Computation time: {0:.3f} sec".format(time.time() - t1))

        # Step3 : GARF
        # output image expressed in counts/samples (generated samples)
        t1 = time.time()
        logger.info(f"Building image with {len(px)}/{current_batch_size} particles")
        garf_param["N_dataset"] = current_batch_size
        img, sq_img = garf.build_arf_image_with_nn(
            garf_nn, garf_model, px, garf_param, verbose=False, debug=debug
        )
        images.append(img)
        sq_images.append(sq_img)
        logger.info("Computation time: {0:.3f} sec".format(time.time() - t1))

        ev += current_batch_size
        pbar.update(current_batch_size)
        ev = min(ev, n)
        logger.info("")

    # mean images
    im_iter = iter(images)
    im = next(im_iter)
    data = itk.GetArrayFromImage(im)
    for im in im_iter:
        d = itk.GetArrayViewFromImage(im)
        data += d
    data = data / len(images)
    img = itk.GetImageFromArray(data)
    img.CopyInformation(images[0])

    # mean images
    im_iter = iter(sq_images)
    im = next(im_iter)
    data = itk.GetArrayFromImage(im)
    for im in im_iter:
        d = itk.GetArrayViewFromImage(im)
        data += d
    data = data / len(sq_images)
    sq_img = itk.GetImageFromArray(data)
    sq_img.CopyInformation(sq_images[0])

    return img, sq_img


def append_gaussian(data, mean, cov, n, vx=None, vy=None):
    x, y = np.random.multivariate_normal(mean, cov, n).T
    d = np.column_stack((x, y))
    if not vx is None:
        d = np.column_stack((d, vx))
    if not vy is None:
        d = np.column_stack((d, vy))
    if data is None:
        return d
    data = np.vstack((data, d))
    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import scipy
import numpy as np
import torch
import gaga_phsp as gaga

speed_of_light = scipy.constants.speed_of_light * 1000 / 1e9


def from_tlor_to_pairs(x, params, verbose=False):
    """
        WARNING: the input 'x' is considered to be a torch Tensor (not numpy)
        Expected options in params: keys_lists, cyl_radius, ignore_directions

        Input:  Cx Cy Cz Vx Vy Vz dAx dAy dAz dBx dBy dBz t1 E1 E2
        Output: t1 t2 X1 Y1 Z1 X2 Y2 Z2 dX1 dY1 dZ1 dX2 dY2 dZ2 E1 E2

    """

    keys = params['keys_list']
    params['keys_output'] = ['t1', 't2', 'X1', 'Y1', 'Z1', 'X2', 'Y2', 'Z2',
                             'dX1', 'dY1', 'dZ1', 'dX2', 'dY2', 'dZ2', 'E1', 'E2']
    # FIXME add weights

    # Step1: name the columns according to key
    C, V, = get_key_3d(x, keys, ['Cx', 'Vx'])
    if params['ignore_directions']:
        params['keys_output'].remove('dX1')
        params['keys_output'].remove('dY1')
        params['keys_output'].remove('dZ1')
        params['keys_output'].remove('dX2')
        params['keys_output'].remove('dY2')
        params['keys_output'].remove('dZ2')
    else:
        dA, dB = get_key_3d(x, keys, ['dAx', 'dBx'])
    tt, E1, E2 = get_key(x, keys, ['t1', 'E1', 'E2'])

    # FIXME weights optional
    w = False
    if 'w1' in keys:
        w = get_key(x, keys, ['w1'])[0]
        params['keys_output'].append('w')

    # Step1: find intersection between line C V and sphere
    A, B, non_valid_index = line_sphere_intersection_torch(params['radius'], C, V)
    if verbose:
        nb_to_remove = torch.unique(non_valid_index, return_counts=True)[1][1]
        print(f'Remove non valid (out of sphere): {nb_to_remove}/{len(A)}')

    # Step2: retrieve time weighted position
    tA, tB = compute_times_wrt_weighted_position(C, A, B, tt)

    # Detect non valid data (negative energy). Keep energy == 0
    non_valid_index = torch.logical_or((E1 <= 0).squeeze(), non_valid_index)
    non_valid_index = torch.logical_or((E2 <= 0).squeeze(), non_valid_index)
    non_valid_index = torch.logical_or((tA <= 0).squeeze(), non_valid_index)
    non_valid_index = torch.logical_or((tB <= 0).squeeze(), non_valid_index)
    if verbose:
        nb_to_remove = torch.unique(non_valid_index, return_counts=True)[1][1]
        print(f'Remove non valid (E<=0): {nb_to_remove}/{len(A)}')

    # Step3: stack
    x = torch.stack((tA, tB), dim=0).T
    x = torch.hstack([x, A])
    x = torch.hstack([x, B])
    if not params['ignore_directions']:
        x = torch.hstack([x, dA])
        x = torch.hstack([x, dB])
    x = torch.hstack([x, E1])
    x = torch.hstack([x, E2])

    # weights
    if 'w1' in keys:
        x = torch.hstack([x, w])

    # mask non valid samples
    x = x[~non_valid_index]

    # end
    return x


def from_pairs_to_tlor(x, params):
    """
        x is numpy array
        Convert pairs into tlor parametrisation
        Input:  t1 t2 X1 Y1 Z1 X2 Y2 Z2 dX1 dY1 dZ1 dX2 dY2 dZ2 E1 E2
        Output: Cx Cy Cz Vx Vy Vz dAx dAy dAz dBx dBy dBz t1 E1 E2 + others
    """

    keys = params['keys_list']
    keys_output = ['Cx', 'Cy', 'Cz',
                   'Vx', 'Vy', 'Vz',
                   'dAx', 'dAy', 'dAz',
                   'dBx', 'dBy', 'dBz',
                   't1', 'E1', 'E2', 'w1']

    # Step1: name the columns according to key
    A, B, dA, dB = get_key_3d(x, keys, ['X1', 'X2', 'dX1', 'dX2'])
    tA, tB, E1, E2 = get_key(x, keys, ['t1', 't2', 'E1', 'E2'])

    # special case for weight that can be ignored
    if 'w1' in keys:
        w1 = get_key(x, keys, ['w1'])[0]
    else:
        w1 = np.ones_like(tA)

    # Step2: compute time weighted position for AB (relative to time at Ap not A)
    # t1 is sum of tA+tB
    C, V, t1 = compute_time_weighted_position(A, B, tA, tB)

    # Step3: stack
    y = np.column_stack([C, V, dA, dB, t1, E1, E2, w1])
    done_keys = ['X1', 'Y1', 'Z1', 'Ax', 'Ay', 'Az',
                 'X2', 'Y2', 'Z2', 'Bx', 'By', 'Bz',
                 'dX1', 'dY1', 'dZ1',
                 'dX2', 'dY2', 'dZ2',
                 't1', 't2', 'E1', 'E2', 'w', 'w1', 'w2']

    # Step4: additional keys
    for k in keys:
        if k not in done_keys:
            z = x[:, keys.index(k)]
            y = np.column_stack([y, z])
            keys_output += [k]

    return y, keys_output


def compute_time_weighted_position(A, B, tA, tB):
    # vector from A to B
    V = (B - A)
    # norm of |AB|
    n = np.linalg.norm(V, axis=1)[:, np.newaxis]
    # relative timing
    tt = tA + tB

    # special case for t = 0 (to avoid divide by zero)
    mask = tt == 0
    tt[mask] = 1.0
    mask = n == 0
    n[mask] = 1
    print(f'Number of total time is 0: {mask.sum()}/{len(tt)}')

    tAr = tA / tt
    tBr = tB / tt
    # relative time weighted position
    CA = A + tAr * V
    CB = B - tBr * V
    # CA should be equal to CB, consider the mean
    C = (CA + CB) / 2
    # normalize direction vector
    V = V / n
    return C, V, tt


def compute_time_at_detector(A, B, Ap, Bp, tA, tB):
    # convert to mm.ns-1
    # c = scipy.constants.speed_of_light * 1000 / 1e9
    # distance from A to Ap
    distance_AAp = np.linalg.norm(Ap - A, axis=1)[:, np.newaxis]
    distance_BBp = np.linalg.norm(Bp - B, axis=1)[:, np.newaxis]
    # convert in time
    tAp = tA + distance_AAp / speed_of_light
    tBp = tB + distance_BBp / speed_of_light
    return tAp, tBp


def line_sphere_intersection_np_old(radius, P, dir):
    # print('line sphere intersection', radius, P.shape, dir.shape)

    print(f'TODO')
    exit()

    # https://en.wikipedia.org/wiki/Line%E2%80%93sphere_intersection
    # nabla ∇
    nabla = np.einsum('ij, ij->i', P, dir)
    nabla = np.square(nabla)
    nabla = nabla - (np.linalg.norm(P, axis=1) ** 2 - radius ** 2)

    # check >0 -> ok
    # print('nabla', nabla)
    # mask = nabla <= 0
    # print('nabla<0', np.count_nonzero(mask))
    non_valid = nabla <= 0

    # distances
    d = -np.einsum('ij, ij->i', P, dir) + np.sqrt(nabla)
    # compute points
    x = P + d[:, np.newaxis] * dir
    return x, non_valid


def line_sphere_intersection_np(radius, P, dir):
    # https://en.wikipedia.org/wiki/Line%E2%80%93sphere_intersection
    a = np.linalg.norm(dir, axis=1, ord=None) ** 2
    b = 2 * np.sum(P * dir, dim=-1)
    # b = 2 * np.sum(P * dir)
    d = np.linalg.norm(P, axis=1, ord=None) ** 2
    c = (d - radius ** 2)
    # delta
    delta = b ** 2 - 4 * a * c
    # consider non valid even if tangent
    non_valid = delta <= 0
    # d
    d1 = (-b - np.sqrt(delta)) / (2 * a)
    d2 = (-b + np.sqrt(delta)) / (2 * a)
    x = P + d1[:, np.newaxis] * dir
    y = P + d2[:, np.newaxis] * dir

    return x, y, non_valid


def line_sphere_intersection_one(radius, P, dir):
    # https://en.wikipedia.org/wiki/Line%E2%80%93sphere_intersection
    a = np.linalg.norm(dir) ** 2
    b = 2 * np.sum(P * dir)
    d = np.linalg.norm(P) ** 2
    c = (d - radius ** 2)
    # delta
    delta = b ** 2 - 4 * a * c
    # consider non valid even if tangent
    non_valid = delta <= 0
    if non_valid:
        return 0, 0, non_valid
    # d
    d1 = (-b - np.sqrt(delta)) / (2 * a)
    d2 = (-b + np.sqrt(delta)) / (2 * a)
    x = P + d1 * dir
    y = P + d2 * dir

    return x, y, non_valid


def line_sphere_intersection_torch(radius, P, dir):
    # https://en.wikipedia.org/wiki/Line%E2%80%93sphere_intersection
    a = torch.linalg.norm(dir, axis=1, ord=None) ** 2
    b = 2 * torch.sum(P * dir, dim=-1)
    d = torch.linalg.norm(P, axis=1, ord=None) ** 2
    c = (d - radius ** 2)
    # delta
    delta = b ** 2 - 4 * a * c
    # consider non valid even if tangent
    non_valid_index = delta <= 0
    # d
    d1 = (-b - torch.sqrt(delta)) / (2 * a)
    d2 = (-b + torch.sqrt(delta)) / (2 * a)
    x = P + d1[:, np.newaxis] * dir
    y = P + d2[:, np.newaxis] * dir

    ## alternative (idem)
    '''dotp = torch.einsum('ij, ij->i', P, dir)
    nabla = torch.square(dotp)
    nabla = nabla - (torch.linalg.norm(P, axis=1) ** 2 - radius ** 2)
    d1 = -dotp - torch.sqrt(nabla)
    d2 = -dotp + torch.sqrt(nabla)
    x = P + d1[:, np.newaxis] * dir
    y = P + d2[:, np.newaxis] * dir'''

    return x, y, non_valid_index


def get_key(x, keys_list, keys):
    v = []
    for k in keys:
        i = keys_list.index(k)
        v.append(x[:, i:i + 1])
    return v


def get_key_3d(x, keys_list, keys):
    # assume Kx Ky kz successive, or X1 Y1 Y1
    v = []
    for k in keys:
        i = keys_list.index(k)
        v.append(x[:, i:i + 3])
    return v


def compute_directions(D, W, t2, t3, A, B):
    """
    Consider the point D and direction W.
        t2 and t3 are the distance from D in W direction
    """
    # retrieve Ap and Bp at virtual spherical detector
    Ap = D - t2 * W
    Bp = D + t3 * W

    # get back the initial A to Ap direction (normalized)
    dA = Ap - A
    n = torch.linalg.norm(dA, axis=1)[:, np.newaxis]
    # FIXME change to torch.nn.functional.normalize ?
    dA = dA / n

    dB = Bp - B
    n = torch.linalg.norm(dB, axis=1)[:, np.newaxis]
    dB = dB / n

    #
    """tAp, tBp = compute_time_at_detector(A, B, Ap, Bp, tA, tB)
    distance_AAp = np.linalg.norm(Ap - A, axis=1)
    distance_BBp = np.linalg.norm(Bp - B, axis=1)
    # convert in time
    tAp = tA + distance_AAp / c
    tBp = tB + distance_BBp / c

    # correct to real timing for A and B: tA and tB
    distance_AAp = np.linalg.norm(Ap - A, axis=1)
    distance_BBp = np.linalg.norm(Bp - B, axis=1)
    # convert in time
    tA = tAp - distance_AAp / c
    tB = tBp - distance_BBp / c
    """

    return dA, dB


def compute_times_wrt_weighted_position(C, A, B, t1):
    """
    According to a point C between AB segment and the time t1,
    compute the times at A and B (tA and tB)
    """
    distance = torch.linalg.norm(B - A, axis=1)
    # timing (t1 is the sum tA+tB)
    t1 = torch.squeeze(t1)
    f = t1 / distance
    # prevent nan -> set both time to zero
    mask = distance == 0
    f[mask] = 0
    # print(f'Number of distance==0 {mask.sum()}/{len(distance)}')
    # distances
    distA = torch.linalg.norm(C - A, axis=1)
    distB = torch.linalg.norm(C - B, axis=1)
    tA = distA * f
    tB = distB * f
    return tA, tB


def line_cylinder_intersections(radius, point, direction):
    """
    Compute intersection between the lines and a cylinder.
    Lines : determine by position 'point', and direction 'direction'
    Cylinder: radius in parameter, height assumed to be Z axis
    """
    # consider xy part
    cx = point[:, 0]
    cy = point[:, 1]
    vx = direction[:, 0]
    vy = direction[:, 1]

    # solve intersection with cylinder
    a = vx ** 2 + vy ** 2
    b = 2 * cx * vx + 2 * cy * vy
    c = cx ** 2 + cy ** 2 - radius ** 2
    delta = b ** 2 - 4 * a * c
    non_valid = delta < 0

    # debug
    print(f'Nb values delta < 0 {torch.count_nonzero(non_valid)}')

    s2 = (-b + torch.sqrt(delta)) / (2 * a)
    s1 = (-b - torch.sqrt(delta)) / (2 * a)
    # print('s1', s1)
    # print('s2', s2)

    # find A and B
    A = point + s1[:, np.newaxis] * direction
    B = point + s2[:, np.newaxis] * direction
    # print('A', A.shape)
    # print('B', B.shape)

    return A, B, non_valid


def plot_sphere_LOR(ax, phsp, keys, x, keys_out, radius):
    # A and B points
    A, B, dA, dB = gaga.get_key_3d(x, keys_out, ['X1', 'X2', 'dX1', 'dX2'])
    ax.plot(A[:, 0], A[:, 1], A[:, 2], '.')
    ax.plot(B[:, 0], B[:, 1], B[:, 2], '.')
    for i in range(len(x)):
        print('A ', A[i])
        print('B', B[i])

    # vectors
    d = B - A
    ax.quiver(A[:, 0], A[:, 1], A[:, 2], d[:, 0], d[:, 1], d[:, 2], arrow_length_ratio=0.1, alpha=0.5)

    # C and V
    C, V, = gaga.get_key_3d(phsp, keys, ['Cx', 'Vx'])
    ax.plot(C[:, 0], C[:, 1], C[:, 2], '.')
    d = 20 * V
    ax.quiver(C[:, 0], C[:, 1], C[:, 2], d[:, 0], d[:, 1], d[:, 2], arrow_length_ratio=0.1, color='b')
    for i in range(len(x)):
        print('C ', C[i])
        print('V', V[i])

    # draw sphere
    u, v = np.mgrid[0:2 * np.pi:20j, 0:np.pi:10j]
    x = np.cos(u) * np.sin(v) * radius
    y = np.sin(u) * np.sin(v) * radius
    z = np.cos(v) * radius
    ax.plot_wireframe(x, y, z, color="r", linewidth=0.1, alpha=0.8)
    # ax.set_aspect("auto")
    ax.set_box_aspect((np.ptp(x), np.ptp(y), np.ptp(z)))


def plot_sphere_pairing(ax, x, keys, radius, type):
    if (type == 'pairs'):
        e1 = x[:, keys.index('E1')]
        mask = e1 != 0
        x = x[mask, :]
        e2 = x[:, keys.index('E2')]
        mask = e2 != 0
        x = x[mask, :]
        plot_sphere_pairing_color(ax, x, keys, radius, 'r')
    if (type == 'singles'):
        e1 = x[:, keys.index('E1')]
        mask = e1 != 0
        x = x[mask, :]
        e2 = x[:, keys.index('E2')]
        mask = e2 == 0
        x = x[mask, :]
        plot_sphere_pairing_color(ax, x, keys, radius, 'g')
    if (type == 'absorbed'):
        e1 = x[:, keys.index('E1')]
        mask = e1 == 0
        x = x[mask, :]
        e2 = x[:, keys.index('E2')]
        mask = e2 == 0
        x = x[mask, :]
        plot_sphere_pairing_color(ax, x, keys, radius, 'b')


def plot_sphere_pairing_color(ax, x, keys_out, radius, color):
    # energy
    e1 = x[:, keys_out.index('E1')]
    e2 = x[:, keys_out.index('E2')]
    t1 = x[:, keys_out.index('t1')]
    t2 = x[:, keys_out.index('t2')]

    # P0
    P0 = gaga.get_key_3d(x, keys_out, ['eX'])[0]
    ax.plot(P0[:, 0], P0[:, 1], P0[:, 2], '.', color=color)
    for i in range(len(P0)):
        ax.text(P0[i, 0], P0[i, 1], P0[i, 2], '%s' % (str(i)), size=8, zorder=1, color='k', alpha=0.7)

    # A and B points
    A, B, dA, dB = gaga.get_key_3d(x, keys_out, ['X1', 'X2', 'dX1', 'dX2'])
    ax.plot(A[:, 0], A[:, 1], A[:, 2], '.')
    ax.plot(B[:, 0], B[:, 1], B[:, 2], '.')
    for i in range(len(A)):
        ax.text(A[i, 0] + 5, A[i, 1] + 5, A[i, 2] + 5, '%s' % (str(i)), size=8, zorder=1, color='k', alpha=0.6)
        ax.text(A[i, 0], A[i, 1], A[i, 2], f'{e1[i]:.2f}', size=8, zorder=1, color='g', alpha=0.6)
        ax.text(B[i, 0], B[i, 1], B[i, 2], f'{e2[i]:.2f}', size=8, zorder=1, color='g', alpha=0.6)
        ax.text(A[i, 0] - 10, A[i, 1] - 10, A[i, 2] - 10, f'{t1[i]:.2f}', size=8, zorder=1, color='b', alpha=0.6)
        ax.text(B[i, 0] - 10, B[i, 1] - 10, B[i, 2] - 10, f'{t2[i]:.2f}', size=8, zorder=1, color='b', alpha=0.6)

    # vectors
    d = B - A
    ax.quiver(A[:, 0], A[:, 1], A[:, 2], d[:, 0], d[:, 1], d[:, 2], arrow_length_ratio=0.01, alpha=0.5)

    # vectors outgoing
    d = dA * 30
    ax.quiver(A[:, 0], A[:, 1], A[:, 2], d[:, 0], d[:, 1], d[:, 2], arrow_length_ratio=0.1, alpha=0.5, color='b')
    d = dB * 30
    ax.quiver(B[:, 0], B[:, 1], B[:, 2], d[:, 0], d[:, 1], d[:, 2], arrow_length_ratio=0.1, alpha=0.5, color='b')

    # draw sphere
    u, v = np.mgrid[0:2 * np.pi:20j, 0:np.pi:10j]
    x = np.cos(u) * np.sin(v) * radius
    y = np.sin(u) * np.sin(v) * radius
    z = np.cos(v) * radius
    ax.plot_wireframe(x, y, z, color="k", linewidth=0.1, alpha=0.8)
    # ax.set_aspect("auto")
    ax.set_box_aspect((np.ptp(x), np.ptp(y), np.ptp(z)))


def pet_pairing_v1(n, i, idx, energy, pos, dir, time, p0, d0, out, nbs):
    if n == 2:
        return pet_pairing_pairs(i, i + idx[1], energy, pos, dir, time, p0, out, nbs)
    if n == 1:
        e1 = energy[i]
        if e1 == 0:
            return pet_pairing_absorbed_v1(i, pos, dir, time, p0, d0, out, nbs)
        else:
            return pet_pairing_singles_v1(i, e1, pos, dir, time, p0, d0, out, nbs)
    nbs.ignored += n


def pet_pairing_v2(n, i, idx, energy, pos, dir, time, p0, d0, out, nbs):
    if n == 2:
        return pet_pairing_pairs(i, i + idx[1], energy, pos, dir, time, p0, out, nbs)
    if n == 1:
        e1 = energy[i]
        if e1 == 0:
            return pet_pairing_absorbed_v2(i, pos, dir, time, p0, d0, out, nbs)
        else:
            return pet_pairing_singles_v2(i, e1, pos, dir, time, p0, d0, out, nbs)
    nbs.ignored += n


def pet_pairing_pairs(idx1, idx2, energy, pos, dir, time, p0, out, nbs):
    out.append([energy[idx1], energy[idx2]] +
               list(pos[idx1, :]) + list(pos[idx2, :]) +
               list(dir[idx1, :]) + list(dir[idx2, :]) +
               [time[idx1], time[idx2]] +
               list(p0[idx1]))
    nbs.pairs += 1


def pet_pairing_absorbed_v1(i, pos, dir, time, p0, d0, out, nbs):
    p0 = p0[i]
    a, b, non_valid = gaga.line_sphere_intersection_one(nbs.radius, p0, d0[i])
    if non_valid:
        return
    t1 = np.linalg.norm(a - p0) / speed_of_light
    t2 = np.linalg.norm(b - p0) / speed_of_light
    out.append([0, 0] +  # energy
               list(a) + list(b) +  # position (unused)
               list(-d0[i]) + list(d0[i]) +  # direction (unused)
               [t1, t2] +  # time (unused)
               list(p0))  # event position
    nbs.absorbed += 1


def pet_pairing_absorbed_v2(i, pos, dir, time, p0, d0, out, nbs):
    p0 = p0[i]
    z = np.zeros_like(p0)
    out.append([0, 0] +  # energy
               list(z) + list(z) +  # position
               list(d0[i]) + list(-d0[i]) +  # direction
               [0, 0] +  # time (unused)
               list(p0))  # event position
    nbs.absorbed += 1


def pet_pairing_singles_v2(i, e1, pos, dir, time, p0, d0, out, nbs):
    p0 = p0[i]
    p1 = pos[i]
    d1 = dir[i]
    t1 = time[i]
    p2 = np.zeros_like(p0)
    out.append([e1, 0] +
               list(p1) + list(p2) +  # p2 position is unused
               list(d1) + list(-d1) +  # d2 direction is unused
               [t1, 0] +  # t2 time is unused
               list(p0))
    nbs.singles += 1


def pet_pairing_singles_v1(i, e1, pos, dir, time, p0, d0, out, nbs):
    p0 = p0[i]
    p1 = pos[i]
    d1 = dir[i]
    t1 = time[i]
    a, b, nv = gaga.line_sphere_intersection_one(nbs.radius, p0, d0[i])
    if nv:
        nbs.ignored += 1
        return
    dia = np.linalg.norm(p1 - a)
    dib = np.linalg.norm(p1 - b)
    # consider the more far
    if dia > dib:
        p2 = a
    else:
        p2 = b
    d2 = p2 - p0
    n2 = np.linalg.norm(d2)
    t2 = n2 / speed_of_light
    d2 = d2 / n2
    out.append([e1, 0] +
               list(p1) + list(p2) +  # p2 position is unused
               list(d1) + list(d2) +  # d2 direction is unused
               [t1, t2] +  # t2 time is unused
               list(p0))
    nbs.singles += 1



import numpy as np
import torch
import gaga_phsp
import datetime
import time
import garf
import gatetools.phsp as phsp
from scipy.stats import kde
from matplotlib import pyplot as plt
from scipy.stats import entropy
from scipy.spatial.transform import Rotation
import logging
import sys


def plot_epoch(ax, params, optim, filename):
    """
    Plot D loss wrt to epoch
    3 panels : all epoch / first 20% / last 1%
    """

    x1 = np.asarray(optim['d_loss_real'])
    x2 = np.asarray(optim['d_loss_fake'])
    # x = -np.add(x1,x2)
    x = -np.asarray(optim['d_loss'])  # with grad penalty

    epoch = np.arange(params['start_epoch'], params['end_epoch'], 1)
    a = ax  # [0]
    l = filename
    a.plot(epoch, x, '-', label='D_loss (GP) ' + l)
    z = np.zeros_like(x)
    a.set_xlabel('epoch')
    a.plot(epoch, z, '--')
    a.legend()

    # print(params['validation_dataset'])
    if not 'validation_dataset' in params or params['validation_dataset'] == None:
        return
    print('validation')
    x = -np.asarray(optim['validation_d_loss'])
    a.plot(epoch, x, '-', label='Valid')
    a.legend()

    return
    # debug
    a = ax[1]
    n = int(len(x) * 0.2)  # first 20%
    xc = x[0:n]
    a.plot(epoch, x, '-', label='D_loss ' + l)
    z = np.zeros_like(xc)
    a.set_xlabel('epoch')
    a.set_xlim((0, n))
    ymin = np.amin(xc)
    ymax = np.amax(xc)
    a.set_ylim((ymin, ymax))
    a.plot(z, '--')
    a.legend()

    a = ax[2]
    n = max(10, int(len(x) * 0.01))  # last 1%
    xc = x
    a.plot(xc, '.-', label='D_loss ' + l)
    z = np.zeros_like(xc)
    a.set_xlabel('epoch')
    a.set_xlim((len(xc) - n, len(xc)))
    xc = x[len(x) - n:len(x)]
    ymin = np.amin(xc)
    ymax = np.amax(xc)
    a.set_ylim((ymin, ymax))
    a.plot(epoch, z, '--')
    a.legend()


def plot_epoch2(ax, params, optim, filename):
    """
    Plot D loss wrt to epoch
    """

    dlr = np.asarray(optim['d_loss_real'])
    dlf = np.asarray(optim['d_loss_fake'])
    dl = np.asarray(optim['d_loss'])  # with grad penalty
    gl = np.asarray(optim['g_loss'])
    epoch = np.arange(params['start_epoch'], params['end_epoch'], 1)

    # one epoch is when all the training dataset is seen
    step = 1  # int(params['training_size'] / params['batch_size'])
    print(step)
    dlr = dlr[::step]
    dlf = dlf[::step]
    dl = dl[::step]
    gl = gl[::step]
    epoch = epoch[::step] / step

    a = ax  # [0]
    l = filename
    # a.plot(epoch, dlr, '-', label='D_loss_real' + l, alpha=0.5)
    # a.plot(epoch, dlf, '-', label='D_loss_fake' + l, alpha=0.5)
    a.plot(epoch, dl, '-', label='D_loss (GP) ' + l)
    a.plot(epoch, gl, '-', label='G_loss ' + l, alpha=0.5)
    z = np.zeros_like(dl)
    a.set_xlabel('epoch')
    a.plot(epoch, z, '--')
    a.legend()

    # print(params['validation_dataset'])
    if not 'validation_dataset' in params or params['validation_dataset'] is None:
        return
    print('Plot with validation dataset')
    x = np.asarray(optim['validation_d_loss'])
    x = x[::step]
    a.plot(epoch, x, '-', label='Valid ' + l)
    a.legend()

    return
    # debug
    a = ax[1]
    n = int(len(x) * 0.2)  # first 20%
    xc = x[0:n]
    a.plot(epoch, x, '-', label='D_loss ' + l)
    z = np.zeros_like(xc)
    a.set_xlabel('epoch')
    a.set_xlim((0, n))
    ymin = np.amin(xc)
    ymax = np.amax(xc)
    a.set_ylim((ymin, ymax))
    a.plot(z, '--')
    a.legend()

    a = ax[2]
    n = max(10, int(len(x) * 0.01))  # last 1%
    xc = x
    a.plot(xc, '.-', label='D_loss ' + l)
    z = np.zeros_like(xc)
    a.set_xlabel('epoch')
    a.set_xlim((len(xc) - n, len(xc)))
    xc = x[len(x) - n:len(x)]
    ymin = np.amin(xc)
    ymax = np.amax(xc)
    a.set_ylim((ymin, ymax))
    a.plot(epoch, z, '--')
    a.legend()


def plot_epoch_wasserstein(ax, optim, filename):
    """
    Plot wasserstein 
    """

    y = np.asarray(optim['w_value'])
    x = np.asarray(optim['w_epoch'])
    if len(x) < 1:
        return

    a = ax[0].twinx()
    a.plot(x, y, '-', color='r', label='W')
    a.legend()

    a = ax[1].twinx()
    a.plot(x, y, '-', color='r', label='W')
    a.legend()

    a = ax[2].twinx()
    a.plot(x, y, '.-', color='r', label='W')
    a.legend()


def fig_plot_marginal(x, k, keys, ax, i, nb_bins, color, r='', lab=''):
    a = phsp.fig_get_sub_fig(ax, i)
    index = keys.index(k)
    if len(x[0]) > 1:
        d = x[:, index]
    else:
        d = x
    # label = ' {} $\mu$={:.2f} $\sigma$={:.2f}'.format(k, np.mean(d), np.std(d))
    label = f'{lab} {k} $\mu$={np.mean(d):.2f} $\sigma$={np.std(d):.2f}'
    if r != '':
        a.hist(d, nb_bins,
               # density=True,
               histtype='stepfilled',
               facecolor=color,
               alpha=0.5,
               range=r,
               label=label)
    else:
        a.hist(d, nb_bins,
               # density=True,
               histtype='stepfilled',
               facecolor=color,
               alpha=0.5,
               label=label)
    a.set_ylabel('Counts')
    a.legend()


def fig_plot_marginal_2d(x, k1, k2, keys, ax, i, nbins, color):
    a = phsp.fig_get_sub_fig(ax, i)
    index1 = keys.index(k1)
    d1 = x[:, index1]
    index2 = keys.index(k2)
    d2 = x[:, index2]
    label = '{} {}'.format(k1, k2)

    ptype = 'scatter'
    # ptype = 'hist'

    if ptype == 'scatter':
        a.scatter(d1, d2, color=color,
                  alpha=0.25,
                  edgecolor='none',
                  s=1)

    if ptype == 'hist':
        cmap = plt.cm.Greens
        if color == 'r':
            cmap = plt.cm.Reds
        a.hist2d(d1, d2,
                 bins=(nbins, nbins),
                 alpha=0.7,
                 cmap=cmap)

    if ptype == 'density':
        x = d1
        y = d2
        print('kde')
        k = kde.gaussian_kde([x, y])
        xi, yi = np.mgrid[x.min():x.max():nbins * 1j, y.min():y.max():nbins * 1j]
        zi = k(np.vstack([xi.flatten(), yi.flatten()]))
        a.pcolormesh(xi, yi,
                     zi.reshape(xi.shape),
                     alpha=0.5)

    a.set_xlabel(k1)
    a.set_ylabel(k2)


def fig_plot_diff_2d(x, y, keys, kk, ax, fig, nb_bins):
    k1 = kk[0]
    k2 = kk[1]
    index1 = keys.index(k1)
    index2 = keys.index(k2)
    x1 = x[:, index1]
    x2 = x[:, index2]
    y1 = y[:, index1]
    y2 = y[:, index2]
    label = '{} {}'.format(k1, k2)

    # compute histo
    H_x, xedges_x, yedges_x = np.histogram2d(x1, x2, bins=nb_bins)
    H_y, xedges_y, yedges_y = np.histogram2d(y1, y2, bins=(xedges_x, yedges_x))

    # make diff
    # H = (H_y - H_x)/H_y
    np.seterr(divide='ignore', invalid='ignore')
    H = np.divide(H_y - H_x, H_y)

    # plot
    X, Y = np.meshgrid(xedges_x, yedges_x)
    im = ax.pcolormesh(X, Y, H.T)  # , norm=LogNorm())
    im.set_clim(vmin=-1, vmax=1)
    fig.colorbar(im, ax=ax)
    ax.set_xlabel(k1)
    ax.set_ylabel(k2)


def fig_plot_projected(data):
    """
    Debug plot
    """

    b = 100
    x = data[:, 0]
    y = data[:, 1]
    theta = data[:, 2]
    phi = data[:, 3]
    E = data[:, 4]

    f, ax = plt.subplots(2, 2, figsize=(10, 10))

    n, bins, patches = ax[0, 0].hist(theta, b, density=True, facecolor='g', alpha=0.35)
    n, bins, patches = ax[0, 1].hist(phi, b, density=True, facecolor='g', alpha=0.35)
    n, bins, patches = ax[1, 0].hist(E * 1000, b, density=True, facecolor='b', alpha=0.35)
    ax[1, 1].scatter(x, y, color='r', alpha=0.35, s=1)

    ax[0, 0].set_xlabel('Theta angle (deg)')
    ax[0, 1].set_xlabel('Phi angle (deg)')
    ax[1, 0].set_xlabel('Energy (keV)')
    ax[1, 1].set_xlabel('X')
    ax[1, 1].set_ylabel('Y')

    plt.tight_layout()
    plt.show()



import setuptools
from setuptools import find_packages

with open("readme.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gaga-phsp",
    version="0.7.0",
    author="David Sarrut",
    author_email="david.sarrut@creatis.insa-lyon.fr",
    description="Python tools for GATE GAN simulations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dsarrut/gaga-phsp",
    packages=find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
    install_requires=[
        "tqdm",
        "colorama",
        "click",
        "scipy",
        "garf>=2.4",
        "matplotlib",
        "gatetools",
        # 'torch'   # the installation of torch is managed by garf
    ],
    scripts=[
        "bin/gaga_train",
        "bin/gaga_info",
        "bin/gaga_plot",
        "bin/gaga_generate",
        "bin/gaga_gauss_test",
        "bin/gaga_gauss_plot",
        "bin/gaga_convert_pth_to_pt",
        "bin/gaga_wasserstein",
        "bin/gaga_garf_generate_img",
        "bin/gaga_pairs_to_tlor",
        "bin/gaga_tlor_to_pairs",
        "bin/gaga_pet_to_pairs_old",
        "bin/gaga_pet_to_pairs",
        "bin/gaga_pet_merge_pairs",
        "bin/gaga_exit_pos_to_ideal_pos",
        "bin/gaga_ideal_pos_to_exit_pos",
    ],
)



import torch.nn as nn
import torch
import gaga_phsp as gaga
from torch import Tensor
from types import MethodType


class MyLeakyReLU(nn.Module):
    def __init__(self, negative_slope=0.01):
        super(MyLeakyReLU, self).__init__()
        self.negative_slope = negative_slope

    def forward(self, x):
        return torch.clamp(x, min=0.0) + torch.clamp(x, max=0.0) * self.negative_slope


def get_activation(params):
    activation = None
    # set activation
    if params['activation'] == 'relu':
        activation = nn.ReLU()
    if params['activation'] == 'leaky_relu':
        activation = nn.LeakyReLU()
    if params['activation'] == 'my_leaky_relu':
        activation = MyLeakyReLU()
    if not activation:
        print('Error, activation unknown: ', params['activation'])
        exit(0)
    return activation


class Discriminator(nn.Module):
    """
    Discriminator: D(x, θD)
    """

    def __init__(self, params):
        super(Discriminator, self).__init__()
        x_dim = params['x_dim']
        d_dim = params['d_dim']
        d_l = params['d_layers']
        sn = False
        if 'spectral_norm' in params:
            sn = params['spectral_norm']
        # activation function
        activation = get_activation(params)
        # create the net
        self.net = nn.Sequential()
        # first layer
        if sn:
            self.net.add_module('1st_layer', nn.utils.spectral_norm(nn.Linear(x_dim, d_dim)))
        else:
            self.net.add_module('1st_layer', nn.Linear(x_dim, d_dim))

        # hidden layers
        for i in range(d_l):
            self.net.add_module(f'activation_{i}', activation)
            if sn:
                self.net.add_module(f'layer_{i}', nn.utils.spectral_norm(nn.Linear(d_dim, d_dim)))
            else:
                self.net.add_module(f'layer_{i}', nn.Linear(d_dim, d_dim))
        # latest layer
        if params['loss'] == 'non-saturating-bce':
            self.net.add_module('sigmoid', nn.Sigmoid())
        else:
            self.net.add_module('last_activation', activation)
            self.net.add_module('last_layer', nn.Linear(d_dim, 1))

        # for p in self.parameters():
        #    if p.ndimension() > 1:
        #        nn.init.kaiming_uniform_(p, nonlinearity='sigmoid')

    def forward(self, x):
        return self.net(x)


class Generator(nn.Module):
    """
    Generator: G(z, θG) -> x fake samples
    """

    def __init__(self, params):
        super(Generator, self).__init__()
        self.params = params
        # the total input dim for the G is z_dim + conditional_keys (if any)
        z_dim = params['z_dim'] + len(params['cond_keys'])
        x_dim = params['x_dim'] - len(params['cond_keys'])
        g_dim = params['g_dim']
        g_l = params['g_layers']
        # activation function
        activation = get_activation(params)
        # create the net
        self.net = nn.Sequential()
        # first layer
        self.net.add_module('first_layer', nn.Linear(z_dim, g_dim))
        # hidden layers
        for i in range(g_l):
            self.net.add_module(f'activation_{i}', activation)
            self.net.add_module(f'layer_{i}', nn.Linear(g_dim, g_dim))
        # last layer
        self.net.add_module(f'last_activation_{i}', activation)
        self.net.add_module(f'last_layer', nn.Linear(g_dim, x_dim))

        # initialisation (not sure better than default init). Keep default.
        for p in self.parameters():
            if p.ndimension() > 1:
                nn.init.kaiming_normal_(p)  ## seems better ???
                # nn.init.xavier_normal_(p)
                # nn.init.kaiming_uniform_(p, nonlinearity='sigmoid')

    def forward(self, x):
        return self.net(x)

    def forward_with_post_processing(self, x):
        # generate data
        y = self.net(x)
        # denormalize
        y = (y * self.x_std) + self.x_mean
        # apply post_process
        y = self.post_process(y, self.params)
        return y

    def init_forward_with_post_processing(self, f, gpu):
        self.post_process = f
        self.init_forward_with_denorm(gpu)
        self.forward = self.forward_with_post_processing

    def init_forward_with_denorm(self, gpu):
        # init the std/mean in torch variable
        device = gaga.init_pytorch_gpu(gpu, False)
        self.x_mean = Tensor(torch.from_numpy(self.params['x_mean']).to(device))
        self.x_std = Tensor(torch.from_numpy(self.params['x_std']).to(device))
        # by default, bypass the test in  denormalization
        # (if forward_with_post_processing is used, the test is kept)
        self.forward = self.forward_with_norm

    def forward_with_norm(self, x):
        y = self.net(x)
        y = (y * self.x_std) + self.x_mean
        return y



import copy
from torch.utils.data import DataLoader
import torch
from torch import Tensor
from tqdm import tqdm
from .gaga_functions import *
from .gaga_helpers import *
import gaga_phsp
from garf.helpers import get_gpu_device

"""
Initial code from :
https://github.com/znxlwm/pytorch-generative-model-collections.git
https://github.com/InsuJeon/Hello-Generative-Model/blob/master/Day04/GAN/gan.ipynb
https://github.com/EmilienDupont/wgan-gp/blob/master/training.py
https://github.com/caogang/wgan-gp/blob/master/gan_toy.py
and others

Disclaimer: experimental work. All mistakes and bullshits are mine.
"""


class Gan(object):
    """
    Main GAN object
    Input params = dict with all parameters and options
    """

    def __init__(self, params):
        # store parameters
        self.params = params

        # init gpu
        self.current_gpu_mode, self.current_gpu_device = get_gpu_device(
            params["gpu_mode"]
        )

        # init model
        self.init_model()

        # init optimiser
        self.init_optimiser()

        # init loss functions
        self.init_loss_functions()

        # init penalty function
        self.init_penalty_functions()

        # init labels
        self.init_labels()

        # compute and store nb of weights
        d_param = filter(lambda p: p.requires_grad, self.D.parameters())
        params["d_nb_weights"] = sum([np.prod(p.size()) for p in d_param])
        g_param = filter(lambda p: p.requires_grad, self.G.parameters())
        params["g_nb_weights"] = sum([np.prod(p.size()) for p in g_param])
        print("Number of parameters for D :", params["d_nb_weights"])
        print("Number of parameters for G :", params["g_nb_weights"])

    def init_model(self):
        """
        Initialise the GAN model, G and D
        """

        p = self.params
        if "start_pth" in p and p["start_pth"] is not None:
            f = p["start_pth"]
            print("Loading previous pth ", f)
            start_params, start_G, start_D, start_optim = gaga_phsp.load(f)
            self.D = start_D
            self.G = start_G
            try:
                self.params["start_epoch"] = start_optim["last_epoch"]
            except:
                self.params["start_epoch"] = start_optim["current_epoch"][-1]
        else:
            self.G, self.D = create_G_and_D_model(p)
            self.params["start_epoch"] = 0

        self.z_rand = get_z_rand(self.params)

    def init_optimiser(self):
        """
        Initialise the optimiser and scheduler
        """

        p = self.params
        d_learning_rate = p["d_learning_rate"]
        g_learning_rate = p["g_learning_rate"]

        if p["optimiser"] == "adam":
            g_weight_decay = float(p["g_weight_decay"])
            d_weight_decay = float(p["d_weight_decay"])
            print("Optimizer regularisation L2 G weight:", g_weight_decay)
            print("Optimizer regularisation L2 D weight:", d_weight_decay)
            beta1 = float(p["beta_1"])
            beta2 = float(p["beta_2"])
            print("Adam beta:", beta1, beta2)
            self.d_optimizer = torch.optim.Adam(
                self.D.parameters(),
                weight_decay=d_weight_decay,
                lr=d_learning_rate,
                betas=(beta1, beta2),
            )
            self.g_optimizer = torch.optim.Adam(
                self.G.parameters(),
                weight_decay=g_weight_decay,
                lr=g_learning_rate,
                betas=(beta1, beta2),
            )

        if p["optimiser"] == "RMSprop":
            # self.d_optimizer = torch.optim.RMSprop(self.D.parameters(), lr=d_learning_rate)
            # self.g_optimizer = torch.optim.RMSprop(self.G.parameters(), lr=g_learning_rate)

            # FIXME
            """
            momentum (float, optional) – momentum factor (default: 0)
            alpha (float, optional) – smoothing constant (default: 0.99)
            centered (bool, optional) – if True, compute the centered RMSProp,
                         the gradient is normalized by an estimation of its variance
            weight_decay (float, optional) – weight decay (L2 penalty) (default: 0)
            """
            self.d_optimizer, self.g_optimizer = get_RMSProp_optimisers(self, p)

        if p["optimiser"] == "SGD":
            self.d_optimizer = torch.optim.SGD(self.D.parameters(), lr=d_learning_rate)
            self.g_optimizer = torch.optim.SGD(self.G.parameters(), lr=g_learning_rate)

        if p["optimiser"] == "gdtuo-RMSprop":
            from gradient_descent_the_ultimate_optimizer import gdtuo

            self.d_optimizer_gdtuo = gdtuo.RMSProp(optimizer=gdtuo.SGD(1e-5))
            self.g_optimizer_gdtuo = gdtuo.RMSProp(optimizer=gdtuo.SGD(1e-5))
            self.d_optimizer = gdtuo.ModuleWrapper(
                self.D, optimizer=self.d_optimizer_gdtuo
            )
            self.g_optimizer = gdtuo.ModuleWrapper(
                self.G, optimizer=self.g_optimizer_gdtuo
            )
            self.d_optimizer.initialize()
            self.g_optimizer.initialize()
            print("NOT YET")
            exit(0)

        # auto decreasing learning_rate
        self.is_scheduler_enabled = False
        try:
            step_size = p["schedule_learning_rate_step"]
            gamma = p["schedule_learning_rate_gamma"]
            print("Scheduler is enabled ", step_size, gamma)
            # WARNING step_size is not nb of epoch but nb of optimiser.step (nb of D update per epoch)
            d_ss = step_size * self.params["d_nb_update"]
            g_ss = step_size * self.params["g_nb_update"]
            print("scheduler: step_size, gamma: ", step_size, gamma)
            print("scheduler d_ss and g_ss:", d_ss, g_ss)
            self.d_scheduler = torch.optim.lr_scheduler.StepLR(
                self.d_optimizer, step_size=d_ss, gamma=gamma
            )
            self.g_scheduler = torch.optim.lr_scheduler.StepLR(
                self.g_optimizer, step_size=g_ss, gamma=gamma
            )
            self.is_scheduler_enabled = True
        except:
            print("Scheduler is disabled")

    def init_loss_functions(self):
        """
        Initialise the loss
        """
        loss = self.params["loss"]
        print(f"Loss is {loss}")

        # https://github.com/AlexiaJM/MaximumMarginGANs/blob/master/Code/GAN.py

        if loss == "wasserstein":
            self.criterion_dr = gaga_phsp.WassersteinNegLoss()
            self.criterion_df = gaga_phsp.WassersteinLoss()
            self.criterion_g = self.criterion_dr
            return

        if loss == "hinge":
            self.criterion_dr = gaga_phsp.HingeNegLoss()
            self.criterion_df = gaga_phsp.HingeLoss()
            self.criterion_g = gaga_phsp.WassersteinNegLoss()
            return

        if loss == "non_saturating_bce":
            self.criterion_dr = nn.BCELoss()
            self.criterion_df = nn.BCELoss()
            return

        print(f"Error, cannot set loss {loss}")
        exit(0)

    def init_penalty_functions(self):
        """
        Initialise the penalty
        """
        t = self.params["penalty"]
        self.penalty_fct = gaga_phsp.zero_penalty
        self.penalty_weight = self.params["penalty_weight"]
        print(f"Penalty weight {self.penalty_weight}")
        print(f"Penalty is: {t}")

        # Change names: 8 different gradient penalties
        # L1_LS L1_Hinge
        # L2_LS L2_Hinge
        # Linf_LS Linf_Hinge
        # GP0 SqHinge
        penalties = {
            "GP_L1_LS": gaga_phsp.GP_L1_LS,
            "GP_L2_LS": gaga_phsp.GP_L2_LS,
            "GP_Linf_LS": gaga_phsp.GP_Linf_LS,
            "GP_L1_Hinge": gaga_phsp.GP_L1_Hinge,
            "GP_L2_Hinge": gaga_phsp.GP_L2_Hinge,
            "GP_Linf_Hinge": gaga_phsp.GP_Linf_Hinge,
            "GP_0GP": gaga_phsp.GP_0GP,
            "GP_SquareHinge": gaga_phsp.GP_SquareHinge,
        }
        # print(t, penalties)
        for p in penalties:
            if t == p:
                self.penalty_fct = penalties[p]
                return

        if t == "clamp":
            self.penalty_fct = gaga_phsp.zero_penalty
            self.clamp_lower = self.params["clamp_lower"]
            self.clamp_upper = self.params["clamp_upper"]
            return

        if t == "no_penalty":
            self.penalty_fct = gaga_phsp.zero_penalty
            return

        print(f"Error, cannot set penalty {t}")
        exit(0)

    def init_labels(self):
        """
        Helper to init the Real=1.0 and Fake=0.0 labels. May be smoothed.
        """
        # Real/Fake labels (1/0)
        batch_size = self.params["batch_size"]
        self.real_labels = Tensor(torch.ones(batch_size, 1)).to(self.current_gpu_device)
        self.fake_labels = Tensor(torch.zeros(batch_size, 1)).to(
            self.current_gpu_device
        )
        # One-sided label smoothing
        if "label_smoothing" in self.params:
            s = self.params["label_smoothing"]
            self.real_labels = Tensor((1.0 - s) + s * torch.rand(batch_size, 1)).to(
                device
            )
            self.fake_labels = Tensor(s * torch.rand(batch_size, 1)).to(
                self.current_gpu_device
            )

    def init_optim_data(self):
        """
        Allocate the optim data structure that store information of the training process
        """

        optim = {}
        optim["g_loss"] = []
        optim["d_loss"] = []
        optim["d_loss_real"] = []
        optim["d_loss_fake"] = []
        optim["g_model_state"] = []
        optim["current_epoch"] = []
        optim["w_value"] = []
        optim["w_epoch"] = []
        optim["validation_d_loss"] = []
        optim["d_best_loss"] = 1e9
        optim["d_best_epoch"] = 0

        return optim

    def set_net_to_device(self, device):
        print("Set model to ", device)
        self.G.to(device)
        self.D.to(device)

        # print('Set data to GPU')
        # real_labels and fake_labels are set to cuda before

        print("Set optim to ", device)
        self.criterion_dr.to(device)
        self.criterion_df.to(device)
        self.criterion_g.to(device)

    def add_Gaussian_noise(self, x, sigma):
        """
        Add Gaussian noise to x. Do nothing is sigma<0
        https://discuss.pytorch.org/t/writing-a-simple-gaussian-noise-layer-in-pytorch/4694/4
        """

        if sigma <= 0:
            return x

        device = torch.device(self.device)

        s = torch.std(x, 0).to(self.current_gpu_device)
        sampled_noise = (
            torch.randn(*x.size(), requires_grad=False).to(self.current_gpu_device)
            * sigma
            * s
        )

        x = x + sampled_noise * sigma
        return x

    # Below : main train function

    def train(self, x):
        """
        Train the GAN
        """

        # normalisation
        print("Normalization")
        x, x_mean, x_std = gaga_phsp.normalize_data(x)
        self.params["x_mean"] = x_mean
        self.params["x_std"] = x_std

        # main dataset
        self.x = x
        self.batch_size = self.params.batch_size

        # init cuda/mps
        self.set_net_to_device(self.current_gpu_device)

        # initialise the data structure that will store info during training
        optim = self.init_optim_data()
        self.optim = optim

        # init conditional
        condn = len(self.params["cond_keys"])
        nx = self.params["x_dim"]
        conditional = condn > 0
        if conditional:
            print(f'Conditional : {self.params["cond_keys"]} ' + str(condn))

        # Sampler
        print("Dataloader")
        batch_size = self.params["batch_size"]
        loader = DataLoader(
            self.x,
            batch_size=batch_size,
            num_workers=2,  # no gain if larger than 2 (?)
            # https://discuss.pytorch.org/t/data-loader-multiprocessing-slow-on-macos/131204/3
            persistent_workers=True,
            pin_memory=True,
            # https://discuss.pytorch.org/t/what-is-the-disadvantage-of-using-pin-memory/1702/4
            shuffle=self.params["shuffle"],
            # shuffle=False,  ## if false ~20% faster, seems identical
            drop_last=True,
        )

        # Start training
        epoch = self.params["start_epoch"]
        self.params["end_epoch"] = epoch + self.params["epoch"]
        print(
            f"Epoch from {self.params['start_epoch']} to {self.params['end_epoch']} (total = {self.params['epoch']})"
        )
        real_labels = self.real_labels
        fake_labels = self.fake_labels
        start = datetime.datetime.now()
        pbar = tqdm(total=self.params["epoch"], disable=not self.params["progress_bar"])
        z_dim = self.params["z_dim"]

        it = iter(loader)

        for batch_idx in range(self.params["epoch"]):
            # Clamp D if needed
            if self.params["penalty"] == "clamp":
                gaga_phsp.clamp_parameters(self)

            # FIXME V3
            for p in self.D.parameters():
                p.requires_grad = True

            # PART 1 : D -------------------------------------------------------
            for _ in range(self.params["d_nb_update"]):
                #
                if self.params["optimiser"] == "gdtuo-RMSprop":
                    self.d_optimizer.begin()
                self.D.zero_grad()

                # the input data
                # https://github.com/pytorch/pytorch/issues/1917
                try:
                    data = next(it)
                except StopIteration:
                    print(
                        "dataset empty, restart from zero", epoch
                    )  # restart from zero
                    it = iter(loader)
                    data = next(it)
                x = Tensor(data).to(self.current_gpu_device)

                ## torch.cuda.empty_cache() ??

                # add instance noise
                x = self.add_Gaussian_noise(x, self.params["r_instance_noise_sigma"])

                # get decision from the discriminator
                d_real_decision = self.D(x)

                # generate z noise (latent)
                z = Tensor(self.z_rand(batch_size, z_dim)).to(self.current_gpu_device)

                # concat conditional vector (if any)
                if conditional:
                    condx = x[:, nx - condn : nx]
                    z = torch.cat((z.float(), condx.float()), dim=1)

                # generate fake data
                # (detach to avoid training G on these labels)
                d_fake_data = self.G(z).detach()  # FIXME detach ?

                # concat conditional vector (if any)
                if conditional:
                    d_fake_data = torch.cat((d_fake_data.float(), condx.float()), dim=1)

                # add instance noise
                d_fake_data = self.add_Gaussian_noise(
                    d_fake_data, self.params["f_instance_noise_sigma"]
                )

                # get the fake decision on the fake data
                d_fake_decision = self.D(d_fake_data)

                # set penalty
                penalty = self.penalty_weight * self.penalty_fct(self, x, d_fake_data)

                # compute loss between decision on real and vector of ones (real_labels)
                d_real_loss = self.criterion_dr(d_real_decision, real_labels)

                # compute loss between decision on fake and vector of zeros (fake_labels)
                d_fake_loss = self.criterion_df(d_fake_decision, fake_labels)

                # backward
                d_real_loss.backward()
                d_fake_loss.backward()
                if self.penalty_fct != gaga_phsp.zero_penalty:
                    penalty.backward()

                # sum of loss
                d_loss = d_real_loss + d_fake_loss + penalty

                # optimizer
                self.d_optimizer.step()

                # scheduler
                if self.is_scheduler_enabled:
                    self.d_scheduler.step()

            # PART 2 : G -------------------------------------------------------
            for p in self.D.parameters():
                p.requires_grad = False

            for _ in range(self.params["g_nb_update"]):
                #
                if self.params["optimiser"] == "gdtuo-RMSprop":
                    self.g_optimizer.begin()

                # required
                self.G.zero_grad()

                # generate z noise (latent)
                z = Tensor(self.z_rand(batch_size, z_dim)).to(self.current_gpu_device)

                # conditional
                if conditional:
                    z = torch.cat((z.float(), condx.float()), dim=1)

                # generate the fake data
                g_fake_data = self.G(z)

                # concat conditional vector (if any)
                if conditional:
                    g_fake_data = torch.cat((g_fake_data.float(), condx.float()), dim=1)

                # add instance noise
                g_fake_data = self.add_Gaussian_noise(
                    g_fake_data, self.params["f_instance_noise_sigma"]
                )

                # get the fake decision
                g_fake_decision = self.D(g_fake_data)

                g_loss = self.criterion_g(g_fake_decision, real_labels)

                # Backprop + Optimize
                g_loss.backward(retain_graph=True)
                self.g_optimizer.step()

                # scheduler
                if self.is_scheduler_enabled:
                    self.g_scheduler.step()

            # housekeeping (to not accumulate gradient)
            # zero_grad clears old gradients from the last step
            # (otherwise you’d just accumulate the gradients from all loss.backward() calls).
            # https://stackoverflow.com/questions/48001598/why-do-we-need-to-call-zero-grad-in-pytorch
            self.D.zero_grad()
            self.G.zero_grad()

            # keep some data (for epoch_dump)
            self.d_loss = d_loss
            self.g_loss = g_loss
            self.d_real_loss = d_real_loss
            self.d_fake_loss = d_fake_loss

            # save loss value
            optim["d_loss_real"].append(d_real_loss.data.item())
            optim["d_loss_fake"].append(d_fake_loss.data.item())
            optim["d_loss"].append(d_loss.data.item())
            optim["g_loss"].append(g_loss.data.item())

            # sometimes: dump, plot, store
            self.epoch_dump(epoch)
            self.epoch_store(epoch)

            # update loop
            pbar.update(1)
            epoch += 1

            # should we stop ?
            if epoch > self.params["end_epoch"]:
                break

        # end of training
        pbar.close()
        stop = datetime.datetime.now()
        optim["last_epoch"] = epoch
        print("Training completed epoch = ", epoch)
        print("Start time    = ", start.strftime(gaga_phsp.date_format))
        print("End time      = ", stop.strftime(gaga_phsp.date_format))
        print("Duration time = ", (stop - start))
        self.params.Duration = str(stop - start)
        self.params.duration = str(stop - start)
        return optim

    def save(self, optim, filename):
        """
        Save the model
        """
        output = dict()
        output["params"] = self.params
        output["optim"] = optim
        state_g = copy.deepcopy(self.G.state_dict())
        state_d = copy.deepcopy(self.D.state_dict())
        output["g_model_state"] = state_g
        output["d_model_state"] = state_d
        torch.save(output, filename)

    def epoch_dump(self, epoch):
        """
        Dump during training
        """
        try:
            n = self.params["epoch_dump"]
        except:
            n = 500

        if epoch % n != 0:
            return
        tqdm.write(
            "Epoch %d d_loss: %.5f   g_loss: %.5f     d_real_loss: %.5f  d_fake_loss: %.5f"
            % (
                epoch,
                self.d_loss.data.item(),
                self.g_loss.data.item(),
                self.d_real_loss.data.item(),
                self.d_fake_loss.data.item(),
            )
        )

    def epoch_store(self, epoch):
        """
        Store during training
        """

        try:
            n = self.params["epoch_store_model_every"]
        except:
            n = -1

        if epoch % n != 0:
            return

        state = copy.deepcopy(self.G.state_dict())
        self.optim["g_model_state"].append(state)
        self.optim["current_epoch"].append(epoch)
